import numpy as np


__all__ = ["LineProfile"]


class LineProfile(object):
    """Horizontal line segment class.

    The line extends `length` arcseconds from
    (`start_x`, `start_y`) at an angle `angle` degrees to the horizontal. Line `width`
    is centered in the perpendicular direction, e.g. a profile with 1 arcsecond width
    and `angle=0` will span -0.5 to 0.5 in the y-direction. Surface brightness is
    constant and given by `amp`.
    """

    param_names = ["amp", "angle", "length", "width", "start_x", "start_y"]
    lower_limit_default = {
        "amp": 0,
        "angle": -180,
        "length": 0.01,
        "width": 0.01,
        "start_x": -100,
        "start_y": -100,
    }
    upper_limit_default = {
        "amp": 10,
        "angle": 180,
        "length": 10,
        "width": 5,
        "start_x": 100,
        "start_y": 100,
    }

    def __init__(self):
        pass

    def function(self, x, y, amp, angle, length, width, start_x=0, start_y=0):
        """Surface brightness per angular unit.

        :param x: x-coordinate on sky
        :param y: y-coordinate on sky
        :param amp: constant surface brightness of line
        :param angle: angle of line to the horizontal (degrees)
        :param length: length of line (arcseconds)
        :param width: width of line (arcseconds), line width extends symmetrically
        :param start_x: ra coordinate of start of line
        :param start_y: dec-coordinate of start of line
        :return: surface brightness, raise as definition is not defined
        """
        ang = -np.deg2rad(angle)
        x_ = np.cos(ang) * (x - start_x) + np.sin(ang) * (y - start_y)
        y_ = np.cos(ang) * (y - start_y) - np.sin(ang) * (x - start_x)
        flux = np.zeros_like(x_)
        flux[(x_ >= 0) * (x_ <= length) * (abs(y_) <= width / 2)] = amp
        return flux

    def total_flux(self, amp, angle, length, width, start_x=0, start_y=0):
        """Integrated flux of the profile.

        :param amp: constant surface brightness of line
        :param angle: angle of line to the horizontal (degrees)
        :param length: length of line (arcseconds)
        :param width: width of line (arcseconds), line width extends symmetrically
        :param start_x: ra coordinate of start of line
        :param start_y: dec-coordinate of start of line
        :return: total flux
        """
        return amp * length * width



__author__ = "dangilman"

import numpy as np
import lenstronomy.Util.param_util as param_util
from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase
from lenstronomy.LensModel.Profiles.epl import EPL
from lenstronomy.LensModel.Profiles.multipole import Multipole, EllipticalMultipole

__all__ = ["EPL_MULTIPOLE_M3M4", "EPL_MULTIPOLE_M3M4_ELL"]


class EPL_MULTIPOLE_M3M4_ELL(LensProfileBase):
    """EPL (Elliptical Power Law) mass profile combined with two elliptical multipole
    terms of order m=3 and m=4 (exact for general axis ratio q).

    See also documentation of EPL_BOXYDIKSY CLASS, lenstronomy.LensModel.Profiles.epl
    and lenstrnomy.LensModel.Profiles.multipole for details.
    """

    param_names = [
        "theta_E",
        "gamma",
        "e1",
        "e2",
        "center_x",
        "center_y",
        "a3_a",
        "delta_phi_m3",
        "a4_a",
        "delta_phi_m4",
    ]
    lower_limit_default = {
        "theta_E": 0,
        "gamma": 1.5,
        "e1": -0.5,
        "e2": -0.5,
        "center_x": -100,
        "center_y": -100,
        "a3_a": -0.2,
        "delta_phi_m3": -np.pi / 6,
        "a4_a": -0.2,
        "delta_phi_m4": -np.pi / 8,
    }
    upper_limit_default = {
        "theta_E": 100,
        "gamma": 2.5,
        "e1": 0.5,
        "e2": 0.5,
        "center_x": 100,
        "center_y": 100,
        "a3_a": 0.2,
        "delta_phi_m3": np.pi / 6,
        "a4_a": 0.2,
        "delta_phi_m4": np.pi / 8,
    }

    def __init__(self):
        self._epl = EPL()
        self._multipole = EllipticalMultipole()
        super(EPL_MULTIPOLE_M3M4_ELL, self).__init__()

    def _param_split(
        self,
        theta_E,
        gamma,
        e1,
        e2,
        a3_a,
        delta_phi_m3,
        a4_a,
        delta_phi_m4,
        center_x=0,
        center_y=0,
    ):
        """This function splits the keyword arguments for the EPL and multipole
        profiles.

        :param theta_E: Einstein radius
        :param gamma: log-slope of EPL mass profile
        :param e1: ellipticity of EPL profile (along 1st axis)
        :param e2: ellipticity of EPL profile (along 2nd axis)
        :param a3_a: amplitude of the m=3 multiple deviation from pure elliptical shape
            related to the physical amplitude of the MULTIPOLE profile by a scaling
            theta_E / sqrt(q)
        :param delta_phi_m3: orientation of the m=3 profile relative to the position
            angle of the EPL profile
        :param a4_a: amplitude of the m=4 multipole deviation from pure elliptical shape
            related to the physical amplitude of the MULTIPOLE profile by a scaling
            theta_E / sqrt(q)
        :param delta_phi_m4: orientation of the m=4 profile relative to the position
            angle of the EPL profile
        :param center_x: center of the profile
        :param center_y: center of the profile
        :return: the keyword arguments for the joint profile
        """

        phi, q = param_util.ellipticity2phi_q(e1, e2)
        rescale_am = theta_E / np.sqrt(q)
        kwargs_epl = {
            "theta_E": theta_E,
            "gamma": gamma,
            "e1": e1,
            "e2": e2,
            "center_x": center_x,
            "center_y": center_y,
        }
        kwargs_multipole_m3 = {
            "m": 3,
            "a_m": a3_a * rescale_am,
            "phi_m": phi + delta_phi_m3,
            "q": q,
            "center_x": center_x,
            "center_y": center_y,
        }
        kwargs_multipole_m4 = {
            "m": 4,
            "a_m": a4_a * rescale_am,
            "phi_m": phi + delta_phi_m4,
            "q": q,
            "center_x": center_x,
            "center_y": center_y,
        }

        return kwargs_epl, kwargs_multipole_m3, kwargs_multipole_m4

    def function(
        self,
        x,
        y,
        theta_E,
        gamma,
        e1,
        e2,
        a3_a,
        delta_phi_m3,
        a4_a,
        delta_phi_m4,
        center_x=0,
        center_y=0,
    ):
        """Computes the gravitational potential in units of theta_E^2.

        :param x: x-coordinate in image plane
        :param y: y-coordinate in image plane
        :param theta_E: Einstein radius
        :param gamma: log-slope of EPL mass profile
        :param e1: ellipticity of EPL profile (along 1st axis)
        :param e2: ellipticity of EPL profile (along 2nd axis)
        :param a3_a: amplitude of the m=3 multiple deviation from pure elliptical shape
            related to the physical amplitude of the MULTIPOLE profile by a scaling
            theta_E / sqrt(q)
        :param delta_phi_m3: orientation of the m=3 profile relative to the position
            angle of the EPL profile
        :param a4_a: amplitude of the m=4 multipole deviation from pure elliptical shape
            related to the physical amplitude of the MULTIPOLE profile by a scaling
            theta_E / sqrt(q)
        :param delta_phi_m4: orientation of the m=4 profile relative to the position
            angle of the EPL profile
        :param center_x: center of the profile
        :param center_y: center of the profile
        :return: lensing potential.
        """
        kwargs_epl, kwargs_multipole3, kwargs_multipole4 = self._param_split(
            theta_E,
            gamma,
            e1,
            e2,
            a3_a,
            delta_phi_m3,
            a4_a,
            delta_phi_m4,
            center_x=center_x,
            center_y=center_y,
        )
        f_epl = self._epl.function(x, y, **kwargs_epl)
        f_multipole = self._multipole.function(x, y, **kwargs_multipole3)
        f_multipole += self._multipole.function(x, y, **kwargs_multipole4)
        return f_epl + f_multipole

    def derivatives(
        self,
        x,
        y,
        theta_E,
        gamma,
        e1,
        e2,
        a3_a,
        delta_phi_m3,
        a4_a,
        delta_phi_m4,
        center_x=0,
        center_y=0,
    ):
        """Computes the derivatives of the potential (deflection angles)in units of
        theta_E.

        :param x: x-coordinate in image plane
        :param y: y-coordinate in image plane
        :param theta_E: Einstein radius
        :param gamma: log-slope of EPL mass profile
        :param e1: ellipticity of EPL profile (along 1st axis)
        :param e2: ellipticity of EPL profile (along 2nd axis)
        :param a3_a: amplitude of the m=3 multiple deviation from pure elliptical shape
            related to the physical amplitude of the MULTIPOLE profile by a scaling
            theta_E / sqrt(q)
        :param delta_phi_m3: orientation of the m=3 profile relative to the position
            angle of the EPL profile
        :param a4_a: amplitude of the m=4 multipole deviation from pure elliptical shape
            related to the physical amplitude of the MULTIPOLE profile by a scaling
            theta_E / sqrt(q)
        :param delta_phi_m4: orientation of the m=4 profile relative to the position
            angle of the EPL profile
        :param center_x: center of the profile
        :param center_y: center of the profile
        :return: alpha_x, alpha_y.
        """
        kwargs_epl, kwargs_multipole3, kwargs_multipole4 = self._param_split(
            theta_E,
            gamma,
            e1,
            e2,
            a3_a,
            delta_phi_m3,
            a4_a,
            delta_phi_m4,
            center_x=center_x,
            center_y=center_y,
        )
        f_x_epl, f_y_epl = self._epl.derivatives(x, y, **kwargs_epl)
        f_x_multipole3, f_y_multipole3 = self._multipole.derivatives(
            x, y, **kwargs_multipole3
        )
        f_x_multipole4, f_y_multipole4 = self._multipole.derivatives(
            x, y, **kwargs_multipole4
        )
        f_x = f_x_epl + f_x_multipole3 + f_x_multipole4
        f_y = f_y_epl + f_y_multipole3 + f_y_multipole4
        return f_x, f_y

    def hessian(
        self,
        x,
        y,
        theta_E,
        gamma,
        e1,
        e2,
        a3_a,
        delta_phi_m3,
        a4_a,
        delta_phi_m4,
        center_x=0,
        center_y=0,
    ):
        """Computes the components of the hessian matrix (second derivatives of the
        potential)

        :param x: x-coordinate in image plane
        :param y: y-coordinate in image plane
        :param theta_E: Einstein radius
        :param gamma: log-slope of EPL mass profile
        :param e1: ellipticity of EPL profile (along 1st axis)
        :param e2: ellipticity of EPL profile (along 2nd axis)
        :param a3_a: amplitude of the m=3 multiple deviation from pure elliptical shape
            related to the physical amplitude of the MULTIPOLE profile by a scaling
            theta_E / sqrt(q)
        :param delta_phi_m3: orientation of the m=3 profile relative to the position
            angle of the EPL profile
        :param a4_a: amplitude of the m=4 multipole deviation from pure elliptical shape
            related to the physical amplitude of the MULTIPOLE profile by a scaling
            theta_E / sqrt(q)
        :param delta_phi_m4: orientation of the m=4 profile relative to the position
            angle of the EPL profile
        :param center_x: center of the profile
        :param center_y: center of the profile
        :return: f_xx, f_xy, f_yx, f_yy.
        """
        kwargs_epl, kwargs_multipole3, kwargs_multipole4 = self._param_split(
            theta_E,
            gamma,
            e1,
            e2,
            a3_a,
            delta_phi_m3,
            a4_a,
            delta_phi_m4,
            center_x=center_x,
            center_y=center_y,
        )
        f_xx_epl, f_xy_epl, f_yx_epl, f_yy_epl = self._epl.hessian(x, y, **kwargs_epl)
        (
            f_xx_multipole3,
            f_xy_multipole3,
            f_yx_multipole3,
            f_yy_multipole3,
        ) = self._multipole.hessian(x, y, **kwargs_multipole3)
        (
            f_xx_multipole4,
            f_xy_multipole4,
            f_yx_multipole4,
            f_yy_multipole4,
        ) = self._multipole.hessian(x, y, **kwargs_multipole4)
        f_xx = f_xx_epl + f_xx_multipole3 + f_xx_multipole4
        f_xy = f_xy_epl + f_xy_multipole3 + f_xy_multipole4
        f_yx = f_yx_epl + f_yx_multipole3 + f_yx_multipole4
        f_yy = f_yy_epl + f_yy_multipole3 + f_yy_multipole4
        return f_xx, f_xy, f_yx, f_yy


class EPL_MULTIPOLE_M3M4(LensProfileBase):
    """EPL (Elliptical Power Law) mass profile combined with two spherical multipole
    terms of order m=3 and m=4 (exact for axis ratio =1).

    Reference to the implementation: https://ui.adsabs.harvard.edu/abs/2022A%26A...659A.127V/abstract

    See also documentation of EPL_BOXYDIKSY CLASS, lenstronomy.LensModel.Profiles.epl and
    lenstrnomy.LensModel.Profiles.multipole for details.
    """

    param_names = [
        "theta_E",
        "gamma",
        "e1",
        "e2",
        "center_x",
        "center_y",
        "a3_a",
        "delta_phi_m3",
        "a4_a",
        "delta_phi_m4",
    ]
    lower_limit_default = {
        "theta_E": 0,
        "gamma": 1.5,
        "e1": -0.5,
        "e2": -0.5,
        "center_x": -100,
        "center_y": -100,
        "a3_a": -0.2,
        "delta_phi_m3": -np.pi / 6,
        "a4_a": -0.2,
        "delta_phi_m4": -np.pi / 8,
    }
    upper_limit_default = {
        "theta_E": 100,
        "gamma": 2.5,
        "e1": 0.5,
        "e2": 0.5,
        "center_x": 100,
        "center_y": 100,
        "a3_a": 0.2,
        "delta_phi_m3": np.pi / 6,
        "a4_a": 0.2,
        "delta_phi_m4": np.pi / 8,
    }

    def __init__(self):
        self._epl = EPL()
        self._multipole = Multipole()
        super(EPL_MULTIPOLE_M3M4, self).__init__()

    def _param_split(
        self,
        theta_E,
        gamma,
        e1,
        e2,
        a3_a,
        delta_phi_m3,
        a4_a,
        delta_phi_m4,
        center_x=0,
        center_y=0,
    ):
        """This function splits the keyword arguments for the EPL and multipole
        profiles.

        :param theta_E: Einstein radius
        :param gamma: log-slope of EPL mass profile
        :param e1: ellipticity of EPL profile (along 1st axis)
        :param e2: ellipticity of EPL profile (along 2nd axis)
        :param a3_a: amplitude of the m=3 multiple deviation from pure elliptical shape
            related to the physical amplitude of the MULTIPOLE profile by a scaling
            theta_E / sqrt(q)
        :param delta_phi_m3: orientation of the m=3 profile relative to the position
            angle of the EPL profile
        :param a4_a: amplitude of the m=4 multipole deviation from pure elliptical shape
            related to the physical amplitude of the MULTIPOLE profile by a scaling
            theta_E / sqrt(q)
        :param delta_phi_m4: orientation of the m=4 profile relative to the position
            angle of the EPL profile
        :param center_x: center of the profile
        :param center_y: center of the profile
        :return: the keyword arguments for the joint profile
        """

        phi, q = param_util.ellipticity2phi_q(e1, e2)
        rescale_am = theta_E / np.sqrt(q)
        kwargs_epl = {
            "theta_E": theta_E,
            "gamma": gamma,
            "e1": e1,
            "e2": e2,
            "center_x": center_x,
            "center_y": center_y,
        }
        kwargs_multipole_m3 = {
            "m": 3,
            "a_m": a3_a * rescale_am,
            "phi_m": phi + delta_phi_m3,
            "center_x": center_x,
            "center_y": center_y,
        }
        kwargs_multipole_m4 = {
            "m": 4,
            "a_m": a4_a * rescale_am,
            "phi_m": phi + delta_phi_m4,
            "center_x": center_x,
            "center_y": center_y,
        }

        return kwargs_epl, kwargs_multipole_m3, kwargs_multipole_m4

    def function(
        self,
        x,
        y,
        theta_E,
        gamma,
        e1,
        e2,
        a3_a,
        delta_phi_m3,
        a4_a,
        delta_phi_m4,
        center_x=0,
        center_y=0,
    ):
        """Computes the gravitational potential in units of theta_E^2.

        :param x: x-coordinate in image plane
        :param y: y-coordinate in image plane
        :param theta_E: Einstein radius
        :param gamma: log-slope of EPL mass profile
        :param e1: ellipticity of EPL profile (along 1st axis)
        :param e2: ellipticity of EPL profile (along 2nd axis)
        :param a3_a: amplitude of the m=3 multiple deviation from pure elliptical shape
            related to the physical amplitude of the MULTIPOLE profile by a scaling
            theta_E / sqrt(q)
        :param delta_phi_m3: orientation of the m=3 profile relative to the position
            angle of the EPL profile
        :param a4_a: amplitude of the m=4 multipole deviation from pure elliptical shape
            related to the physical amplitude of the MULTIPOLE profile by a scaling
            theta_E / sqrt(q)
        :param delta_phi_m4: orientation of the m=4 profile relative to the position
            angle of the EPL profile
        :param center_x: center of the profile
        :param center_y: center of the profile
        :return: lensing potential.
        """
        kwargs_epl, kwargs_multipole3, kwargs_multipole4 = self._param_split(
            theta_E,
            gamma,
            e1,
            e2,
            a3_a,
            delta_phi_m3,
            a4_a,
            delta_phi_m4,
            center_x=center_x,
            center_y=center_y,
        )
        f_epl = self._epl.function(x, y, **kwargs_epl)
        f_multipole = self._multipole.function(x, y, **kwargs_multipole3)
        f_multipole += self._multipole.function(x, y, **kwargs_multipole4)
        return f_epl + f_multipole

    def derivatives(
        self,
        x,
        y,
        theta_E,
        gamma,
        e1,
        e2,
        a3_a,
        delta_phi_m3,
        a4_a,
        delta_phi_m4,
        center_x=0,
        center_y=0,
    ):
        """Computes the derivatives of the potential (deflection angles)in units of
        theta_E.

        :param x: x-coordinate in image plane
        :param y: y-coordinate in image plane
        :param theta_E: Einstein radius
        :param gamma: log-slope of EPL mass profile
        :param e1: ellipticity of EPL profile (along 1st axis)
        :param e2: ellipticity of EPL profile (along 2nd axis)
        :param a3_a: amplitude of the m=3 multiple deviation from pure elliptical shape
            related to the physical amplitude of the MULTIPOLE profile by a scaling
            theta_E / sqrt(q)
        :param delta_phi_m3: orientation of the m=3 profile relative to the position
            angle of the EPL profile
        :param a4_a: amplitude of the m=4 multipole deviation from pure elliptical shape
            related to the physical amplitude of the MULTIPOLE profile by a scaling
            theta_E / sqrt(q)
        :param delta_phi_m4: orientation of the m=4 profile relative to the position
            angle of the EPL profile
        :param center_x: center of the profile
        :param center_y: center of the profile
        :return: alpha_x, alpha_y.
        """
        kwargs_epl, kwargs_multipole3, kwargs_multipole4 = self._param_split(
            theta_E,
            gamma,
            e1,
            e2,
            a3_a,
            delta_phi_m3,
            a4_a,
            delta_phi_m4,
            center_x=center_x,
            center_y=center_y,
        )
        f_x_epl, f_y_epl = self._epl.derivatives(x, y, **kwargs_epl)
        f_x_multipole3, f_y_multipole3 = self._multipole.derivatives(
            x, y, **kwargs_multipole3
        )
        f_x_multipole4, f_y_multipole4 = self._multipole.derivatives(
            x, y, **kwargs_multipole4
        )
        f_x = f_x_epl + f_x_multipole3 + f_x_multipole4
        f_y = f_y_epl + f_y_multipole3 + f_y_multipole4
        return f_x, f_y

    def hessian(
        self,
        x,
        y,
        theta_E,
        gamma,
        e1,
        e2,
        a3_a,
        delta_phi_m3,
        a4_a,
        delta_phi_m4,
        center_x=0,
        center_y=0,
    ):
        """Computes the components of the hessian matrix (second derivatives of the
        potential)

        :param x: x-coordinate in image plane
        :param y: y-coordinate in image plane
        :param theta_E: Einstein radius
        :param gamma: log-slope of EPL mass profile
        :param e1: ellipticity of EPL profile (along 1st axis)
        :param e2: ellipticity of EPL profile (along 2nd axis)
        :param a3_a: amplitude of the m=3 multiple deviation from pure elliptical shape
            related to the physical amplitude of the MULTIPOLE profile by a scaling
            theta_E / sqrt(q)
        :param delta_phi_m3: orientation of the m=3 profile relative to the position
            angle of the EPL profile
        :param a4_a: amplitude of the m=4 multipole deviation from pure elliptical shape
            related to the physical amplitude of the MULTIPOLE profile by a scaling
            theta_E / sqrt(q)
        :param delta_phi_m4: orientation of the m=4 profile relative to the position
            angle of the EPL profile
        :param center_x: center of the profile
        :param center_y: center of the profile
        :return: f_xx, f_xy, f_yx, f_yy.
        """
        kwargs_epl, kwargs_multipole3, kwargs_multipole4 = self._param_split(
            theta_E,
            gamma,
            e1,
            e2,
            a3_a,
            delta_phi_m3,
            a4_a,
            delta_phi_m4,
            center_x=center_x,
            center_y=center_y,
        )
        f_xx_epl, f_xy_epl, f_yx_epl, f_yy_epl = self._epl.hessian(x, y, **kwargs_epl)
        (
            f_xx_multipole3,
            f_xy_multipole3,
            f_yx_multipole3,
            f_yy_multipole3,
        ) = self._multipole.hessian(x, y, **kwargs_multipole3)
        (
            f_xx_multipole4,
            f_xy_multipole4,
            f_yx_multipole4,
            f_yy_multipole4,
        ) = self._multipole.hessian(x, y, **kwargs_multipole4)
        f_xx = f_xx_epl + f_xx_multipole3 + f_xx_multipole4
        f_xy = f_xy_epl + f_xy_multipole3 + f_xy_multipole4
        f_yx = f_yx_epl + f_yx_multipole3 + f_yx_multipole4
        f_yy = f_yy_epl + f_yy_multipole3 + f_yy_multipole4
        return f_xx, f_xy, f_yx, f_yy



import pytest
from lenstronomy.GalKin.analytic_kinematics import AnalyticKinematics
import numpy as np
from lenstronomy.GalKin.numeric_kinematics import NumericKinematics
from astropy.cosmology import FlatLambdaCDM
import numpy.testing as npt


class TestAnalyticKinematics(object):
    def setup_method(self):
        pass

    def test_sigma_s2(self):
        kwargs_aperture = {
            "center_ra": 0,
            "width": 1,
            "length": 1,
            "angle": 0,
            "center_dec": 0,
            "aperture_type": "slit",
        }
        kwargs_cosmo = {"d_d": 1000, "d_s": 1500, "d_ds": 800}
        kwargs_psf = {"psf_type": "GAUSSIAN", "fwhm": 1}
        kin = AnalyticKinematics(
            kwargs_cosmo,
            interpol_grid_num=2000,
            log_integration=True,
            max_integrate=100,
            min_integrate=5e-6,
        )
        kwargs_light = {"r_eff": 1}
        sigma_s2 = kin.sigma_s2(
            r=1,
            R=0.1,
            kwargs_mass={"theta_E": 1, "gamma": 2},
            kwargs_light=kwargs_light,
            kwargs_anisotropy={"r_ani": 1},
        )
        npt.assert_almost_equal(sigma_s2[0], 70885880558.5913, decimal=3)

    def test_properties(self):
        kwargs_aperture = {
            "center_ra": 0,
            "width": 1,
            "length": 1,
            "angle": 0,
            "center_dec": 0,
            "aperture_type": "slit",
        }
        kwargs_cosmo = {"d_d": 1000, "d_s": 1500, "d_ds": 800}
        kwargs_psf = {"psf_type": "GAUSSIAN", "fwhm": 1}
        kin = AnalyticKinematics(
            kwargs_cosmo,
            interpol_grid_num=2000,
            log_integration=True,
            max_integrate=150,
            min_integrate=5e-6,
        )

        assert kin.max_integrate == 150
        assert kin.min_integrate == 5e-6

    def test_draw_light(self):
        kin = AnalyticKinematics
        assert kin._get_hernquist_scale_radius({"a": 1}) == 1
        assert kin._get_hernquist_scale_radius({"Rs": 2}) == 2
        assert kin._get_hernquist_scale_radius({"r_eff": 4}) == 4 * 0.551

        with pytest.raises(ValueError):
            kin._get_hernquist_scale_radius({"not_Rs": 1})

    def test_I_R_sigma2_and_IR(self):
        kwargs_aperture = {
            "center_ra": 0,
            "width": 1,
            "length": 1,
            "angle": 0,
            "center_dec": 0,
            "aperture_type": "slit",
        }
        kwargs_cosmo = {"d_d": 1000, "d_s": 1500, "d_ds": 800}
        kwargs_psf = {"psf_type": "GAUSSIAN", "fwhm": 1}
        kin = AnalyticKinematics(
            kwargs_cosmo,
            interpol_grid_num=10000,
            log_integration=False,
            max_integrate=100,
            min_integrate=1e-4,
        )
        kwargs_mass = {"theta_E": 1, "gamma": 2}
        kwargs_light = {"r_eff": 1}
        kwargs_ani = {"r_ani": 1}
        IR_sigma2, IR = kin._I_R_sigma2(
            R=1,
            kwargs_mass=kwargs_mass,
            kwargs_light=kwargs_light,
            kwargs_anisotropy=kwargs_ani,
        )

        kin._log_int = True
        kin._interp_grid_num = 1000
        IR_sigma2_2, IR_2 = kin._I_R_sigma2(
            R=1,
            kwargs_mass=kwargs_mass,
            kwargs_light=kwargs_light,
            kwargs_anisotropy=kwargs_ani,
        )
        assert IR_sigma2 - IR_sigma2_2 < 10

    def test_against_numeric_profile(self):
        z_d = 0.295
        z_s = 0.657

        kwargs_model = {
            "mass_profile_list": ["EPL"],
            "light_profile_list": ["HERNQUIST"],
            "anisotropy_model": "OM",
            # 'lens_redshift_list': [z_d],
        }

        cosmo = FlatLambdaCDM(H0=70, Om0=0.3)
        D_d = cosmo.angular_diameter_distance(z_d).value
        D_s = cosmo.angular_diameter_distance(z_s).value
        D_ds = cosmo.angular_diameter_distance_z1z2(z_d, z_s).value

        kwargs_cosmo = {"d_d": D_d, "d_s": D_s, "d_ds": D_ds}

        numeric_kin = NumericKinematics(
            kwargs_model,
            kwargs_cosmo,
            interpol_grid_num=1000,
            max_integrate=1000,
            min_integrate=1e-4,
        )
        analytic_kin = AnalyticKinematics(
            kwargs_cosmo,
            interpol_grid_num=2000,
            log_integration=True,
            max_integrate=100,
            min_integrate=1e-4,
        )

        R = np.logspace(-5, np.log10(6), 100)
        r_eff = 1.85
        theta_e = 1.63
        gamma = 2
        a_ani = 1

        numeric_s2ir, numeric_ir = numeric_kin.I_R_sigma2_and_IR(
            R,
            [{"theta_E": theta_e, "gamma": gamma, "center_x": 0.0, "center_y": 0.0}],
            [{"Rs": r_eff * 0.551, "amp": 1.0, "center_x": 0.0, "center_y": 0.0}],
            {"r_ani": a_ani * r_eff},
        )

        numeric_vel_dis = np.sqrt(numeric_s2ir / numeric_ir) / 1e3

        analytic_s2ir = np.zeros_like(R)
        analytic_ir = np.zeros_like(R)

        for i, r in enumerate(R):
            analytic_s2ir[i], analytic_ir[i] = analytic_kin.I_R_sigma2_and_IR(
                r,
                {"theta_E": theta_e, "gamma": gamma, "center_x": 0.0, "center_y": 0.0},
                {"Rs": r_eff * 0.551, "amp": 1.0, "center_x": 0.0, "center_y": 0.0},
                {"r_ani": a_ani * r_eff},
            )

        analytic_vel_dis = np.sqrt(analytic_s2ir / analytic_ir) / 1e3

        # check if matches below 1%
        npt.assert_array_less(
            (numeric_vel_dis - analytic_vel_dis) / numeric_vel_dis,
            0.01 * np.ones_like(numeric_vel_dis),
        )


if __name__ == "__main__":
    pytest.main()



import pytest
import numpy as np
import numpy.testing as npt
import unittest
import warnings

from lenstronomy.Data.psf import PSF
import lenstronomy.Util.kernel_util as kernel_util
import lenstronomy.Util.image_util as image_util


class TestData(object):
    def setup_method(self):
        self.deltaPix = 0.05
        fwhm = 0.2
        kwargs_gaussian = {
            "psf_type": "GAUSSIAN",
            "fwhm": fwhm,
            "truncation": 5,
            "pixel_size": self.deltaPix,
        }
        self.psf_gaussian = PSF(**kwargs_gaussian)
        kernel_point_source = kernel_util.kernel_gaussian(
            num_pix=21, delta_pix=self.deltaPix, fwhm=fwhm
        )
        kwargs_pixel = {"psf_type": "PIXEL", "kernel_point_source": kernel_point_source}
        self.psf_pixel = PSF(**kwargs_pixel)

    def test_kernel_point_source(self):
        kernel_gaussian = self.psf_gaussian.kernel_point_source
        kernel_pixel = self.psf_pixel.kernel_point_source
        assert len(kernel_gaussian) == 21
        assert len(kernel_pixel) == 21

        kwargs_psf = {
            "psf_type": "GAUSSIAN",
            "fwhm": 0.2,
            "truncation": 3,
            "pixel_size": 0.05,
        }
        psf_class = PSF(**kwargs_psf)
        kernel_point_source = psf_class.kernel_point_source
        assert len(kernel_point_source) == 13
        kernel_super = psf_class.kernel_point_source_supersampled(
            supersampling_factor=3
        )
        npt.assert_almost_equal(
            np.sum(kernel_point_source), np.sum(kernel_super), decimal=9
        )
        npt.assert_almost_equal(np.sum(kernel_point_source), 1, decimal=9)

    def test_kernel_subsampled(self):
        deltaPix = 0.05  # pixel size of image
        numPix = 40  # number of pixels per axis
        subsampling_res = 3  # subsampling scale factor (in each dimension)
        fwhm = 0.3  # FWHM of the PSF kernel
        fwhm_object = 0.2  # FWHM of the Gaussian source to be convolved

        # create Gaussian/Pixelized kernels
        # first we create the sub-sampled kernel
        kernel_point_source_subsampled = kernel_util.kernel_gaussian(
            num_pix=11 * subsampling_res,
            delta_pix=deltaPix / subsampling_res,
            fwhm=fwhm,
        )
        # to have the same consistent kernel, we re-size (average over the sub-sampled pixels) the sub-sampled kernel
        kernel_point_source = image_util.re_size(
            kernel_point_source_subsampled, subsampling_res
        )
        # here we create the two PSF() classes
        kwargs_pixel_subsampled = {
            "psf_type": "PIXEL",
            "kernel_point_source": kernel_point_source_subsampled,
            "point_source_supersampling_factor": subsampling_res,
        }
        psf_pixel_subsampled = PSF(**kwargs_pixel_subsampled)
        psf_pixel_subsampled.kernel_point_source_supersampled(
            supersampling_factor=subsampling_res + 1
        )
        kernel_point_source /= np.sum(kernel_point_source)
        kwargs_pixel = {"psf_type": "PIXEL", "kernel_point_source": kernel_point_source}
        psf_pixel = PSF(**kwargs_pixel)

        kernel_point_source = psf_pixel.kernel_point_source
        kernel_super = psf_pixel.kernel_point_source_supersampled(
            supersampling_factor=3
        )
        npt.assert_almost_equal(
            np.sum(kernel_point_source), np.sum(kernel_super), decimal=8
        )
        npt.assert_almost_equal(np.sum(kernel_point_source), 1, decimal=8)

        deltaPix = 0.05  # pixel size of image
        numPix = 40  # number of pixels per axis
        subsampling_res = 4  # subsampling scale factor (in each dimension)
        fwhm = 0.3  # FWHM of the PSF kernel
        fwhm_object = 0.2  # FWHM of the Gaussian source to be convolved

        # create Gaussian/Pixelized kernels
        # first we create the sub-sampled kernel
        kernel_point_source_subsampled = kernel_util.kernel_gaussian(
            num_pix=11 * subsampling_res + 1,
            delta_pix=deltaPix / subsampling_res,
            fwhm=fwhm,
        )

        kwargs_pixel_subsampled = {
            "psf_type": "PIXEL",
            "kernel_point_source": kernel_point_source_subsampled,
            "point_source_supersampling_factor": subsampling_res,
        }
        psf_pixel_subsampled = PSF(**kwargs_pixel_subsampled)
        kernel_point_source /= np.sum(kernel_point_source)
        kwargs_pixel = {"psf_type": "PIXEL", "kernel_point_source": kernel_point_source}
        psf_pixel = PSF(**kwargs_pixel)
        kernel_point_source = psf_pixel.kernel_point_source
        kernel_point_source_new = psf_pixel_subsampled.kernel_point_source
        npt.assert_almost_equal(
            np.sum(kernel_point_source), np.sum(kernel_point_source_new), decimal=8
        )
        npt.assert_almost_equal(np.sum(kernel_point_source), 1, decimal=8)

        psf_none = PSF(psf_type="NONE")
        kernel_super = psf_none.kernel_point_source_supersampled(supersampling_factor=5)
        npt.assert_almost_equal(kernel_super, psf_none.kernel_point_source, decimal=9)

    def test_supersampling_repeated(self):
        # test to retrieve the original supersampled kernel despite in the intermediate steps using
        # a different supersampling factor
        subsampling_res = 5
        deltaPix = 0.05  # pixel size of image
        fwhm = 0.3  # FWHM of the PSF kernel

        kernel_point_source_subsampled = kernel_util.kernel_gaussian(
            num_pix=11 * subsampling_res,
            delta_pix=deltaPix / subsampling_res,
            fwhm=fwhm,
        )

        kwargs_pixel_subsampled = {
            "psf_type": "PIXEL",
            "kernel_point_source": kernel_point_source_subsampled,
            "point_source_supersampling_factor": subsampling_res,
            "kernel_point_source_normalisation": False,
        }
        psf = PSF(**kwargs_pixel_subsampled)
        kernel_supersampled_new = psf.kernel_point_source_supersampled(
            supersampling_factor=subsampling_res - 1, updata_cache=True
        )
        kernel_supersampled_origin = psf.kernel_point_source_supersampled(
            supersampling_factor=subsampling_res
        )
        npt.assert_almost_equal(
            kernel_supersampled_origin, kernel_point_source_subsampled
        )

    def test_fwhm(self):
        deltaPix = 1.0
        fwhm = 5.6
        kwargs = {
            "psf_type": "GAUSSIAN",
            "fwhm": fwhm,
            "truncation": 5,
            "pixel_size": deltaPix,
        }
        psf_kernel = PSF(**kwargs)
        fwhm_compute = psf_kernel.fwhm
        assert fwhm_compute == fwhm

        kernel = kernel_util.kernel_gaussian(num_pix=31, delta_pix=deltaPix, fwhm=fwhm)
        kwargs = {
            "psf_type": "PIXEL",
            "truncation": 5,
            "pixel_size": deltaPix,
            "kernel_point_source": kernel,
        }
        psf_kernel = PSF(**kwargs)
        fwhm_compute = psf_kernel.fwhm
        npt.assert_almost_equal(fwhm_compute, fwhm, decimal=1)

        kwargs = {
            "psf_type": "PIXEL",
            "truncation": 5,
            "pixel_size": deltaPix,
            "kernel_point_source": kernel,
            "point_source_supersampling_factor": 1,
        }
        psf_kernel = PSF(**kwargs)
        fwhm_compute = psf_kernel.fwhm
        npt.assert_almost_equal(fwhm_compute, fwhm, decimal=1)

    def test_kernel_pixel(self):
        deltaPix = 1.0
        fwhm = 5.6
        kwargs = {
            "psf_type": "GAUSSIAN",
            "fwhm": fwhm,
            "truncation": 5,
            "pixel_size": deltaPix,
        }
        psf_kernel = PSF(**kwargs)
        kernel_pixel = psf_kernel.kernel_pixel
        npt.assert_almost_equal(
            np.sum(kernel_pixel), np.sum(psf_kernel.kernel_point_source), decimal=9
        )

    def test_psf_error_map(self):
        deltaPix = 1.0
        fwhm = 5.6
        kwargs = {
            "psf_type": "GAUSSIAN",
            "fwhm": fwhm,
            "truncation": 5,
            "pixel_size": deltaPix,
        }
        psf_kernel = PSF(**kwargs)
        error_map = psf_kernel.psf_error_map
        assert error_map.all() == 0

    def test_warning(self):
        deltaPix = 0.05  # pixel size of image
        subsampling_res = 4  # subsampling scale factor (in each dimension)
        fwhm = 0.3  # FWHM of the PSF kernel

        # create Gaussian/Pixelized kernels
        # first we create the sub-sampled kernel
        kernel_point_source_subsampled = kernel_util.kernel_gaussian(
            num_pix=11 * subsampling_res + 1,
            delta_pix=deltaPix / subsampling_res,
            fwhm=fwhm,
        )
        print(len(kernel_point_source_subsampled), "test")
        kwargs_psf = {
            "psf_type": "PIXEL",
            "kernel_point_source": kernel_point_source_subsampled,
            "point_source_supersampling_factor": subsampling_res,
            "psf_error_map": np.ones_like(kernel_point_source_subsampled),
        }
        psf_kernel = PSF(**kwargs_psf)
        n = len(psf_kernel.kernel_point_source)
        error_map = psf_kernel.psf_error_map
        assert len(error_map) == n

    def test_unnormalized(self):
        psf_norm_factor = 0.1
        kernel_point_source = np.zeros((51, 51))
        kernel_point_source[25, 25] = psf_norm_factor

        kwargs_psf = {
            "psf_type": "PIXEL",
            "kernel_point_source": kernel_point_source,
            "point_source_supersampling_factor": 1,
            "kernel_point_source_normalisation": False,
        }
        psf = PSF(**kwargs_psf)
        npt.assert_almost_equal(np.sum(psf.kernel_point_source), psf_norm_factor)
        npt.assert_almost_equal(np.sum(psf.kernel_pixel), psf_norm_factor)
        npt.assert_almost_equal(
            np.sum(psf.kernel_point_source_supersampled(supersampling_factor=1)),
            psf_norm_factor,
        )

        kwargs_psf = {
            "psf_type": "PIXEL",
            "kernel_point_source": kernel_point_source,
            "point_source_supersampling_factor": 5,
            "kernel_point_source_normalisation": False,
        }
        psf = PSF(**kwargs_psf)
        npt.assert_almost_equal(np.sum(psf.kernel_point_source), psf_norm_factor)
        npt.assert_almost_equal(np.sum(psf.kernel_pixel), psf_norm_factor)
        npt.assert_almost_equal(
            np.sum(psf.kernel_point_source_supersampled(supersampling_factor=5)),
            psf_norm_factor,
        )

        kwargs_psf = {
            "psf_type": "PIXEL",
            "kernel_point_source": kernel_point_source,
            "point_source_supersampling_factor": 1,
            "kernel_point_source_normalisation": True,
        }
        psf = PSF(**kwargs_psf)
        npt.assert_almost_equal(np.sum(psf.kernel_point_source), 1)
        npt.assert_almost_equal(np.sum(psf.kernel_pixel), 1)
        npt.assert_almost_equal(
            np.sum(psf.kernel_point_source_supersampled(supersampling_factor=1)), 1
        )

        kwargs_psf = {
            "psf_type": "PIXEL",
            "kernel_point_source": kernel_point_source,
            "point_source_supersampling_factor": 5,
            "kernel_point_source_normalisation": True,
        }
        psf = PSF(**kwargs_psf)
        npt.assert_almost_equal(np.sum(psf.kernel_point_source), 1)
        npt.assert_almost_equal(np.sum(psf.kernel_pixel), 1)
        npt.assert_almost_equal(
            np.sum(psf.kernel_point_source_supersampled(supersampling_factor=5)), 1
        )


class TestRaise(unittest.TestCase):
    def test_raise(self):
        psf = PSF(psf_type="PIXEL", kernel_point_source=np.ones((3, 3)))
        psf.psf_type = "WRONG"
        with self.assertRaises(ValueError):
            PSF(psf_type="GAUSSIAN")
        with self.assertRaises(ValueError):
            PSF(psf_type="PIXEL")
        with self.assertRaises(ValueError):
            PSF(psf_type="PIXEL", kernel_point_source=np.ones((2, 2)))
        with self.assertRaises(ValueError):
            PSF(psf_type="WRONG")
        with self.assertRaises(ValueError):
            PSF(
                psf_type="PIXEL",
                kernel_point_source=np.ones((3, 3)),
                psf_error_map=np.ones((5, 5)),
            )
            psf.kernel_point_source_supersampled(supersampling_factor=3)
        with self.assertRaises(ValueError):
            psf = PSF(psf_type="PIXEL", kernel_point_source=np.ones((3, 3)))
            psf.psf_type = "WRONG"
            psf.kernel_point_source_supersampled(supersampling_factor=3)
        with self.assertRaises(ValueError):
            psf = PSF(psf_type="GAUSSIAN", fwhm=100, pixel_size=0.0001)
            psf.kernel_point_source_supersampled(supersampling_factor=3)

        with warnings.catch_warnings(record=True) as w:
            # Cause all warnings to always be triggered.
            warnings.simplefilter("always")
            # Trigger a warning.
            kernel_point_source_subsampled = np.ones((9, 9))
            subsampling_res = 3
            kwargs_pixel_subsampled = {
                "psf_type": "PIXEL",
                "kernel_point_source": kernel_point_source_subsampled,
                "point_source_supersampling_factor": subsampling_res,
            }
            psf_pixel_subsampled = PSF(**kwargs_pixel_subsampled)
            psf_pixel_subsampled.kernel_point_source_supersampled(
                supersampling_factor=subsampling_res + 4
            )
            # Verify some things
            assert 1 == 1
            # assert len(w) == 1
            # assert issubclass(w[-1].category, Warning)


if __name__ == "__main__":
    pytest.main()



import numpy.testing as npt
import numpy as np


class TestTracerModelFit(object):
    def setup_method(self):
        # imagng data specifics
        background_rms = 0.005  # background noise per pixel
        exp_time = 500.0  # exposure time (arbitrary units, flux per pixel is in units #photons/exp_time unit)
        numPix = 60  # cutout pixel size per axis
        pixel_scale = 0.05  # pixel size in arcsec (area per pixel = pixel_scale**2)
        fwhm = 0.05  # full width at half maximum of PSF
        psf_type = "GAUSSIAN"  # 'GAUSSIAN', 'PIXEL', 'NONE'

        # tracer measurements specifics
        tracer_noise_map = np.ones(
            (numPix, numPix)
        )  # variance of metallicity measurement for each pixel

        # lensing quantities
        lens_model_list = ["SIE", "SHEAR"]
        kwargs_sie = {
            "theta_E": 0.66,
            "center_x": 0.05,
            "center_y": 0,
            "e1": 0.07,
            "e2": -0.03,
        }  # parameters of the deflector lens model
        kwargs_shear = {
            "gamma1": 0.0,
            "gamma2": -0.05,
        }  # shear values to the source plane

        kwargs_lens = [kwargs_sie, kwargs_shear]
        from lenstronomy.LensModel.lens_model import LensModel

        lens_model_class = LensModel(lens_model_list)

        # Sersic parameters in the initial simulation for the source
        kwargs_sersic = {
            "amp": 16,
            "R_sersic": 0.1,
            "n_sersic": 1,
            "e1": -0.1,
            "e2": 0.1,
            "center_x": 0.1,
            "center_y": 0,
        }
        source_model_list = ["SERSIC_ELLIPSE"]
        kwargs_source = [kwargs_sersic]

        from lenstronomy.LightModel.light_model import LightModel

        source_model_class = LightModel(source_model_list)

        # Source tracer model
        tracer_source_model_list = ["LINEAR"]
        kwargs_tracer_source = [
            {
                "amp": 100,
                "k": 1,
                "center_x": kwargs_sersic["center_x"],
                "center_y": kwargs_sersic["center_y"],
            }
        ]
        tracer_source_class = LightModel(tracer_source_model_list)

        # lens light model
        kwargs_sersic_lens = {
            "amp": 16,
            "R_sersic": 0.6,
            "n_sersic": 2,
            "e1": -0.1,
            "e2": 0.1,
            "center_x": 0.05,
            "center_y": 0,
        }

        lens_light_model_list = ["SERSIC_ELLIPSE"]
        kwargs_lens_light = [kwargs_sersic_lens]
        lens_light_model_class = LightModel(lens_light_model_list)

        # import main simulation class of lenstronomy
        from lenstronomy.Util import util
        from lenstronomy.Data.imaging_data import ImageData
        from lenstronomy.Data.psf import PSF
        import lenstronomy.Util.image_util as image_util
        from lenstronomy.ImSim.image_model import ImageModel

        # generate the coordinate grid and image properties (we only read out the relevant lines we need)
        (
            _,
            _,
            ra_at_xy_0,
            dec_at_xy_0,
            _,
            _,
            Mpix2coord,
            _,
        ) = util.make_grid_with_coordtransform(
            numPix=numPix,
            deltapix=pixel_scale,
            center_ra=0,
            center_dec=0,
            subgrid_res=1,
            inverse=False,
        )

        kwargs_image_data = {
            "background_rms": background_rms,  # rms of background noise
            "exposure_time": exp_time,  # exposure time (or a map per pixel)
            "ra_at_xy_0": ra_at_xy_0,  # RA at (0,0) pixel
            "dec_at_xy_0": dec_at_xy_0,  # DEC at (0,0) pixel
            "transform_pix2angle": Mpix2coord,
            # matrix to translate shift in pixel in shift in relative RA/DEC (2x2 matrix). Make sure it's units are arcseconds or the angular units you want to model.
            "image_data": np.zeros((numPix, numPix)),
            # 2d data vector, here initialized with zeros as place holders that get's overwritten once a simulated image with noise is created.
        }

        image_data_class = ImageData(**kwargs_image_data)
        # generate the psf variables
        kwargs_psf = {
            "psf_type": "GAUSSIAN",
            "fwhm": fwhm,
            "pixel_size": pixel_scale,
            "truncation": 3,
        }

        # if you are using a PSF estimate from e.g. a star in the FoV of your exposure, you can set
        # kwargs_psf = {'psf_type': 'PIXEL', 'pixel_size': deltaPix, 'kernel_point_source': 'odd numbered 2d grid with centered star/PSF model'}

        psf_class = PSF(**kwargs_psf)
        kwargs_numerics = {
            "supersampling_factor": 1,
            "supersampling_convolution": False,
        }

        imageModel = ImageModel(
            image_data_class,
            psf_class,
            lens_model_class=lens_model_class,
            source_model_class=source_model_class,
            lens_light_model_class=lens_light_model_class,
            kwargs_numerics=kwargs_numerics,
        )

        # generate image
        image_model = imageModel.image(
            kwargs_lens,
            kwargs_source,
            kwargs_lens_light=kwargs_lens_light,
            kwargs_ps=None,
        )

        poisson = image_util.add_poisson(image_model, exp_time=exp_time)
        bkg = image_util.add_background(image_model, sigma_bkd=background_rms)
        image_real = image_model + poisson + bkg

        image_data_class.update_data(image_real)
        kwargs_image_data["image_data"] = image_real

        # Tracer data

        kwargs_tracer_data = {
            "noise_map": tracer_noise_map,  # variance of pixels
            "ra_at_xy_0": ra_at_xy_0,  # RA at (0,0) pixel
            "dec_at_xy_0": dec_at_xy_0,  # DEC at (0,0) pixel
            "transform_pix2angle": Mpix2coord,
            # matrix to translate shift in pixel in shift in relative RA/DEC (2x2 matrix). Make sure it's units are arcseconds or the angular units you want to model.
            "image_data": np.zeros((numPix, numPix)),
            # 2d data vector, here initialized with zeros as place holders that get's overwritten once a simulated image with noise is created.
        }

        tracer_data_class = ImageData(**kwargs_tracer_data)

        from lenstronomy.ImSim.tracer_model import TracerModelSource

        tracer_model_class = TracerModelSource(
            data_class=tracer_data_class,
            tracer_source_class=tracer_source_class,
            psf_class=psf_class,
            lens_model_class=lens_model_class,
            source_model_class=source_model_class,
            lens_light_model_class=lens_light_model_class,
        )

        tracer_model = tracer_model_class.tracer_model(
            kwargs_tracer_source, kwargs_lens, kwargs_source
        )

        # add noise
        tracer_noise = image_util.add_background(
            image_model, sigma_bkd=tracer_noise_map
        )
        tracer_real = tracer_model + tracer_noise

        tracer_data_class.update_data(tracer_real)
        kwargs_tracer_data["image_data"] = tracer_real

        # model fitting

        # lens models
        fixed_lens = []
        kwargs_lens_init = []
        kwargs_lens_sigma = []
        kwargs_lower_lens = []
        kwargs_upper_lens = []

        fixed_lens.append(
            {}
        )  # for this example, we fix the power-law index of the lens model to be isothermal
        kwargs_lens_init.append(
            {"theta_E": 0.7, "e1": 0.0, "e2": 0.0, "center_x": 0.0, "center_y": 0.0}
        )
        kwargs_lens_sigma.append(
            {"theta_E": 0.2, "e1": 0.05, "e2": 0.05, "center_x": 0.05, "center_y": 0.05}
        )
        kwargs_lower_lens.append(
            {"theta_E": 0.01, "e1": -0.5, "e2": -0.5, "center_x": -10, "center_y": -10}
        )
        kwargs_upper_lens.append(
            {"theta_E": 10.0, "e1": 0.5, "e2": 0.5, "center_x": 10, "center_y": 10}
        )

        fixed_lens.append({"ra_0": 0, "dec_0": 0})
        kwargs_lens_init.append({"gamma1": 0.0, "gamma2": 0.0})
        kwargs_lens_sigma.append({"gamma1": 0.1, "gamma2": 0.1})
        kwargs_lower_lens.append({"gamma1": -0.2, "gamma2": -0.2})
        kwargs_upper_lens.append({"gamma1": 0.2, "gamma2": 0.2})

        lens_params = [
            kwargs_lens_init,
            kwargs_lens_sigma,
            fixed_lens,
            kwargs_lower_lens,
            kwargs_upper_lens,
        ]

        fixed_source = []
        kwargs_source_init = []
        kwargs_source_sigma = []
        kwargs_lower_source = []
        kwargs_upper_source = []

        fixed_source.append({})
        kwargs_source_init.append(
            {
                "R_sersic": 0.2,
                "n_sersic": 1,
                "e1": 0,
                "e2": 0,
                "center_x": 0.0,
                "center_y": 0,
                "amp": 16,
            }
        )
        kwargs_source_sigma.append(
            {
                "n_sersic": 0.5,
                "R_sersic": 0.1,
                "e1": 0.05,
                "e2": 0.05,
                "center_x": 0.2,
                "center_y": 0.2,
                "amp": 10,
            }
        )
        kwargs_lower_source.append(
            {
                "e1": -0.5,
                "e2": -0.5,
                "R_sersic": 0.001,
                "n_sersic": 0.5,
                "center_x": -10,
                "center_y": -10,
                "amp": 0,
            }
        )
        kwargs_upper_source.append(
            {
                "e1": 0.5,
                "e2": 0.5,
                "R_sersic": 10,
                "n_sersic": 5.0,
                "center_x": 10,
                "center_y": 10,
                "amp": 100,
            }
        )

        source_params = [
            kwargs_source_init,
            kwargs_source_sigma,
            fixed_source,
            kwargs_lower_source,
            kwargs_upper_source,
        ]

        fixed_lens_light = []
        kwargs_lens_light_init = []
        kwargs_lens_light_sigma = []
        kwargs_lower_lens_light = []
        kwargs_upper_lens_light = []

        fixed_lens_light.append({})
        kwargs_lens_light_init.append(
            {
                "R_sersic": 0.5,
                "n_sersic": 2,
                "e1": 0,
                "e2": 0,
                "center_x": 0.0,
                "center_y": 0,
                "amp": 16,
            }
        )
        kwargs_lens_light_sigma.append(
            {
                "n_sersic": 1,
                "R_sersic": 0.3,
                "e1": 0.05,
                "e2": 0.05,
                "center_x": 0.1,
                "center_y": 0.1,
                "amp": 10,
            }
        )
        kwargs_lower_lens_light.append(
            {
                "e1": -0.5,
                "e2": -0.5,
                "R_sersic": 0.001,
                "n_sersic": 0.5,
                "center_x": -10,
                "center_y": -10,
                "amp": 0,
            }
        )
        kwargs_upper_lens_light.append(
            {
                "e1": 0.5,
                "e2": 0.5,
                "R_sersic": 10,
                "n_sersic": 5.0,
                "center_x": 10,
                "center_y": 10,
                "amp": 100,
            }
        )

        lens_light_params = [
            kwargs_lens_light_init,
            kwargs_lens_light_sigma,
            fixed_lens_light,
            kwargs_lower_lens_light,
            kwargs_upper_lens_light,
        ]

        # Tracer parameter configurations

        fixed_tracer_source = [{}]
        kwargs_tracer_source_init = [
            {"amp": 100, "k": 1, "center_x": 0.0, "center_y": 0}
        ]
        kwargs_tracer_source_sigma = [
            {"amp": 50, "k": 1, "center_x": 0.2, "center_y": 0.2}
        ]
        kwargs_lower_tracer_source = [
            {"amp": 0, "k": -2, "center_x": -5, "center_y": -5}
        ]
        kwargs_upper_tracer_source = [
            {"amp": 1000, "k": 5, "center_x": 5, "center_y": 5}
        ]

        tracer_source_params = [
            kwargs_tracer_source_init,
            kwargs_tracer_source_sigma,
            fixed_tracer_source,
            kwargs_lower_tracer_source,
            kwargs_upper_tracer_source,
        ]

        kwargs_params = {
            "lens_model": lens_params,
            "source_model": source_params,
            "lens_light_model": lens_light_params,
            "tracer_source_model": tracer_source_params,
        }

        kwargs_likelihood = {"source_marg": False}
        kwargs_model = {
            "lens_model_list": lens_model_list,
            "source_light_model_list": source_model_list,
            "lens_light_model_list": lens_light_model_list,
            "tracer_source_model_list": tracer_source_model_list,
            "tracer_source_band": 0,
        }  # what imaging band's surface brightness solution is used for tracer models

        multi_band_list = [[kwargs_image_data, kwargs_psf, kwargs_numerics]]
        # if you have multiple  bands to be modeled simultaneously, you can append them to the mutli_band_list
        kwargs_data_joint = {
            "multi_band_list": multi_band_list,
            "multi_band_type": "single-band",
            # 'multi-linear': every imaging band has independent solutions of the surface brightness, 'joint-linear': there is one joint solution of the linear coefficients demanded across the bands.
            "tracer_data": [kwargs_tracer_data, kwargs_psf, kwargs_numerics],
        }
        kwargs_constraints = {
            "linear_solver": True,
            # optional, if 'linear_solver': False, lenstronomy does not apply a linear inversion of the 'amp' parameters during fitting but instead samples them.
            "joint_source_light_with_tracer": [[0, 0, ["center_x", "center_y"]]],
        }  # link source light with tracer [[i_source_light, k_tracer_source, ['param_name1', 'param_name2', ...]], [...], ...]
        self.kwargs_data_joint = kwargs_data_joint
        self.kwargs_model = kwargs_model
        self.kwargs_constraints = kwargs_constraints
        self.kwargs_likelihood = kwargs_likelihood
        self.kwargs_params = kwargs_params

    def test_run_fit(self):
        from lenstronomy.Workflow.fitting_sequence import FittingSequence

        fitting_seq = FittingSequence(
            self.kwargs_data_joint,
            self.kwargs_model,
            self.kwargs_constraints,
            self.kwargs_likelihood,
            self.kwargs_params,
        )

        fitting_kwargs_list = [
            [
                "update_settings",
                {
                    "tracer_source_add_fixed": [
                        [0, ["amp", "k"]]
                    ],  # add tracer fixed to avoid sampling
                    "kwargs_likelihood": {
                        "tracer_likelihood": False
                    },  # remove tracer data to be fitted
                },
            ],
            ["PSO", {"sigma_scale": 1.0, "n_particles": 10, "n_iterations": 10}],
            [
                "update_settings",
                {
                    "tracer_source_remove_fixed": [
                        [0, ["amp", "k"]]
                    ],  # remove tracer fixed to sample
                    "kwargs_likelihood": {"tracer_likelihood": True},
                },
            ],  # evaluate tracer data likelihood
            ["PSO", {"sigma_scale": 0.1, "n_particles": 10, "n_iterations": 10}],
            # ['MCMC', {'n_burn': 10, 'n_run': 10, 'n_walkers': 10, 'sigma_scale': .1}]
        ]

        chain_list = fitting_seq.fit_sequence(fitting_kwargs_list)
        kwargs_result = fitting_seq.best_fit()
        npt.assert_almost_equal(
            kwargs_result["kwargs_tracer_source"][0]["k"], 1, decimal=1
        )



from lenstronomy.LightModel.Profiles.lineprofile import LineProfile
import numpy as np
import numpy.testing as npt


class TestLineProfile(object):
    def setup_method(self):
        self.lineprofile = LineProfile()

    def test_function(self):
        amp = 1
        length = 1
        width = 0.01
        angle = 57
        x = np.array([0, 1, 0.5 * np.cos(np.deg2rad(-angle))])
        y = np.array([0, 0, 0.5 * np.sin(np.deg2rad(-angle))])
        x_single = 0.2 * np.cos(np.deg2rad(-angle))
        y_single = 0.2 * np.sin(np.deg2rad(-angle))
        flux_true = np.array([amp, 0, amp])
        flux = self.lineprofile.function(x, y, amp, angle, length, width)
        single_flux_true = amp
        single_flux = self.lineprofile.function(
            x_single, y_single, amp, angle, length, width
        )
        npt.assert_equal(flux_true, flux)
        npt.assert_equal(single_flux_true, single_flux)

    def test_total_flux(self):
        amp = 1
        length = 1
        width = 0.01
        angle = 57
        total_flux = self.lineprofile.total_flux(amp, angle, length, width)
        total_flux_true = length * width * amp
        npt.assert_equal(total_flux_true, total_flux)



__author__ = "ajshajib", "dgilman", "sibirrer"

import numpy as np
from scipy.integrate import quad
from scipy.special import hyp2f1
from scipy.interpolate import interp1d
from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase

__all__ = ["GNFW"]


class GNFW(LensProfileBase):
    """
    This class computes the lensing quantities of a generalized NFW profile:

    .. math::
        \\rho(r) = \\frac{\\rho_{\\rm s}} { (r/r_{\\rm s}})^{\\gamma_{\\rm in}} * (1 + r/r_{\\rm
        s})^{3 - {\\gamma_{\\rm in}}}

    This class uses the normalization parameter `kappa_s` defined as:

    .. math::
        kappas_{\\rm s} = \\frac{\\rho_{\\rm s} r_{\\rm s}}{\\Sigma_{\\rm crit}}

    Some expressions are obtained from Keeton 2001
    https://ui.adsabs.harvard.edu/abs/2001astro.ph..2341K/abstract. See and cite the
    references therein.
    """

    model_name = "GNFW"
    _s = 0.001  # numerical limit for minimal radius
    param_names = ["Rs", "alpha_Rs", "gamma_in", "center_x", "center_y"]
    lower_limit_default = {
        "Rs": 0,
        "alpha_Rs": 0,
        "gamma_in": 0.0,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "Rs": 100,
        "alpha_Rs": 10.0,
        "gamma_in": 3.0,
        "center_x": 100,
        "center_y": 100,
    }

    def __init__(self, trapezoidal_integration=False, integration_steps=1000):
        """

        :param trapezoidal_integrate: bool, if True, the numerical integral is performed
         with the trapezoidal rule, otherwise with ~scipy.integrate.quad
        :param integration_steps: number of steps in the trapezoidal integral
        """
        super(GNFW, self).__init__()
        self._integration_steps = integration_steps
        if trapezoidal_integration:
            self._integrate = self._trapezoidal_integrate
        else:
            self._integrate = self._quad_integrate

        gamma_ins = np.linspace(0.0, 2.99, 300)
        alphas = np.zeros_like(gamma_ins)
        for i, g in enumerate(gamma_ins):
            alphas[i] = self._alpha(1.0, 1.0, 1.0, gamma_in=g)
        self._alpha_1_interp = interp1d(gamma_ins, alphas, kind="cubic")

    def function(self, x, y, Rs, alpha_Rs, gamma_in, center_x=0, center_y=0):
        """Potential of gNFW profile.

        :param x: angular position
        :type x: float/numpy array
        :param y: angular position
        :type y: float/numpy array
        :param Rs: angular turn over point
        :type Rs: float
        :param alpha_Rs: deflection (angular units) at projected Rs
        :type alpha_Rs: float
        :param gamma_in: inner slope
        :type gamma_in: float
        :param center_x: center of halo
        :type center_x: float
        :param center_y: center of halo
        :type center_y: float
        :return: potential at radius r
        :rtype: float
        """
        x_ = x - center_x
        y_ = y - center_y
        r = np.sqrt(x_**2 + y_**2)
        r = np.maximum(r, self._s)

        if Rs < 0.0000001:
            Rs = 0.0000001

        kappa_s = self.alpha_Rs_to_kappa_s(Rs, alpha_Rs, gamma_in)

        if isinstance(r, int) or isinstance(r, float):
            return self._num_integral_potential(r, Rs, kappa_s, gamma_in)
        else:
            # TODO: currently the numerical integral is done one by one. More efficient is sorting the radial list and
            # then perform one numerical integral reading out to the radial points
            f_ = []
            for _r in r:
                f_.append(self._num_integral_potential(_r, Rs, kappa_s, gamma_in))
            return np.array(f_)

    def _num_integral_potential(self, r, Rs, kappa_s, gamma_in):
        """Compute the numerical integral of the potential.

        :param r: radius of interest
        :type r: float
        :param Rs: scale radius
        :type Rs: float
        :param kappa_s: convergence at Rs
        :type kappa_s: float
        :type alpha_Rs: float
        :param gamma_in: inner slope
        :type gamma_in: float
        :return: potential at radius r
        :rtype: float
        """

        def _integrand(x):
            return self._alpha(x, Rs, kappa_s, gamma_in)

        return quad(_integrand, a=0, b=r)[0]

    def derivatives(self, x, y, Rs, alpha_Rs, gamma_in, center_x=0, center_y=0):
        """Returns df/dx and df/dy of the function.

        :param x: angular position
        :type x: float/numpy array
        :param y: angular position
        :type y: float/numpy array
        :param Rs: angular turn over point
        :type Rs: float
        :param alpha_Rs: deflection (angular units) at projected Rs
        :type alpha_Rs: float
        :param gamma_in: inner slope
        :type gamma_in: float
        :param center_x: center of halo
        :type center_x: float
        :param center_y: center of halo
        :type center_y: float
        :return: deflection angle in x, deflection angle in y
        :rtype: float, float
        """
        if Rs < 0.0000001:
            Rs = 0.0000001
        x_ = x - center_x
        y_ = y - center_y
        R = np.sqrt(x_**2 + y_**2)
        R = np.maximum(R, self._s)

        f_r = self.alpha(R, Rs, alpha_Rs, gamma_in)
        f_x = f_r * x_ / R
        f_y = f_r * y_ / R

        return f_x, f_y

    def hessian(self, x, y, Rs, alpha_Rs, gamma_in, center_x=0, center_y=0):
        """Returns Hessian matrix of function d^2f/dx^2, d^f/dy^2, d^2/dxdy.

        :param x: angular position
        :type x: float/numpy array
        :param y: angular position
        :type y: float/numpy array
        :param Rs: angular turn over point
        :type Rs: float
        :param alpha_Rs: deflection (angular units) at projected Rs
        :type alpha_Rs: float
        :param gamma_in: inner slope
        :type gamma_in: float
        :param center_x: center of halo
        :type center_x: float
        :param center_y: center of halo
        :type center_y: float
        :return: f_xx, f_xy, f_xy, f_yy
        :rtype: float, float, float, float
        """
        if Rs < 0.0000001:
            Rs = 0.0000001
        x_ = x - center_x
        y_ = y - center_y
        R = np.sqrt(x_**2 + y_**2)

        kappa = self.kappa(R, Rs, alpha_Rs, gamma_in)
        f_r = self.alpha(R, Rs, alpha_Rs, gamma_in)
        f_rr = 2 * kappa - f_r / R

        cos_t = x_ / R
        sin_t = y_ / R

        f_xx = cos_t**2 * f_rr + sin_t**2 / R * f_r
        f_yy = sin_t**2 * f_rr + cos_t**2 / R * f_r
        f_xy = cos_t * sin_t * f_rr - cos_t * sin_t / R * f_r

        return f_xx, f_xy, f_xy, f_yy

    def density(self, R, Rs, rho0, gamma_in):
        """Three dimensional generalized NFW profile.

        :param R: radius of interest
        :type R: float/numpy array
        :param Rs: scale radius
        :type Rs: float
        :param rho0: density normalization
        :type rho0: float
        :param gamma_in: inner slope
        :type gamma_in: float
        :return: rho(R) density
        :rtype: float
        """
        return rho0 * (R / Rs) ** -gamma_in * (1 + R / Rs) ** (gamma_in - 3)

    def density_lens(self, R, Rs, alpha_Rs, gamma_in):
        """Computes the density at 3d radius r given lens model parameterization. The
        integral in the LOS projection of this quantity results in the convergence
        quantity.

        :param R: radius of interest
        :type R: float/numpy array
        :param Rs: scale radius
        :type Rs: float
        :param alpha_Rs: deflection at Rs
        :type alpha_Rs: float
        :param gamma_in: inner slope
        :type gamma_in: float
        :return: density at radius R
        :rtype: float
        """
        rho0 = self.alpha2rho0(alpha_Rs, Rs, gamma_in)
        return self.density(R, Rs, rho0, gamma_in)

    def density_2d(self, x, y, Rs, rho0, gamma_in, center_x=0, center_y=0):
        """Projected two dimenstional NFW profile (kappa*Sigma_crit)

        :param x: x-coordinate
        :type x: float/numpy array
        :param y: y-coordinate
        :type y: float/numpy array
        :param Rs: scale radius
        :type Rs: float
        :param rho0: density normalization (characteristic density)
        :type rho0: float
        :param gamma_in: inner slope
        :type gamma_in: float
        :param center_x: center of halo
        :type center_x: float
        :param center_y: center of halo
        :type center_y: float
        :return: Epsilon(R) projected density at radius R
        """
        x_ = x - center_x
        y_ = y - center_y
        R = np.sqrt(x_**2 + y_**2)
        kappa_s = self.rho0_to_kappa_s(rho0, Rs)

        return self._kappa(R, Rs, kappa_s, gamma_in)

    def mass_3d(self, R, Rs, rho0, gamma_in):
        """Mass enclosed a 3d sphere or radius r.

        :param R: radius of interest
        :type R: float/numpy array
        :param Rs: scale radius
        :type Rs: float
        :param rho0: density normalization (characteristic density)
        :type rho0: float
        :param gamma_in: inner slope
        :type gamma_in: float
        :return: mass enclosed a 3d sphere or radius r
        :rtype: float
        """
        M_0 = 4 * np.pi * rho0 * Rs**3 / (3 - gamma_in)
        x = R / Rs
        return (
            M_0
            * x ** (3 - gamma_in)
            * hyp2f1(3 - gamma_in, 3 - gamma_in, 4 - gamma_in, -x)
        )

    def mass_3d_lens(self, R, Rs, alpha_Rs, gamma_in):
        """Mass enclosed a 3d sphere or radius r given a lens parameterization with
        angular units.

        :param R: radius of interest
        :type R: float/numpy array
        :param Rs: scale radius
        :type Rs: float
        :param alpha_Rs: deflection at Rs
        :type alpha_Rs: float
        :param gamma_in: inner slope
        :type gamma_in: float
        :return: mass enclosed a 3d sphere or radius r
        :rtype: float
        """
        rho0 = self.alpha2rho0(alpha_Rs, Rs, gamma_in)
        return self.mass_3d(R, Rs, rho0, gamma_in)

    def _trapezoidal_integrate(self, func, x, gamma_in):
        """Integrate a function using the trapezoid rule.

        :param func: function to integrate
        :type func: function
        :param x: x = R/Rs
        :type x: float
        :param gamma_in: inner slope
        :type gamma_in: float
        :param steps: number of steps
        :type steps: int
        :return: integral
        :rtype: float
        """
        steps = self._integration_steps
        y = np.linspace(1e-10, 1 - 1e-10, steps)
        dy = y[1] - y[0]

        weights = np.ones(steps)
        weights[0] = 0.5
        weights[-1] = 0.5

        if isinstance(x, int) or isinstance(x, float):
            integral = np.sum(func(y, x, gamma_in) * dy * weights)
        else:
            x_flat = x.flatten()
            ys = np.repeat(y[:, np.newaxis], len([x_flat]), axis=1)

            integral = np.sum(
                func(ys, x_flat, gamma_in) * dy * weights[:, np.newaxis], axis=0
            )
            integral = integral.reshape(x.shape)

        return integral

    def _quad_integrate(self, func, x, gamma_in):
        """Integrate a function using the trapezoid rule.

        :param func: function to integrate
        :type func: function
        :param x: x = R/Rs
        :type x: float
        :param gamma_in: inner slope
        :type gamma_in: float
        :param steps: number of steps
        :type steps: int
        :return: integral
        :rtype: float
        """
        if isinstance(x, int) or isinstance(x, float):
            integral = quad(func, a=0, b=1, args=(x, gamma_in))[0]
        else:
            integral = np.zeros_like(x)

            for i in range(len(x)):
                integral[i] = quad(func, a=0, b=1, args=(x[i], gamma_in))[0]

        return integral

    def _alpha_integrand(self, y, x, gamma_in):
        """Integrand of the deflection angel integral.

        :param y: integration variable
        :type y: np.array
        :param x: x = R/Rs
        :type x: float
        :param gamma_in: inner slope
        :type gamma_in: float
        :return: integrand of the deflection angel integral
        """
        return (y + x) ** (gamma_in - 3) * (1 - np.sqrt(1 - y**2)) / y

    def _kappa_integrand(self, y, x, gamma_in):
        """Integrand of the deflection angel integral in eq. (57) of Keeton 2001.

        :param y: integration variable
        :type y: np.array
        :param x: x = R/Rs
        :type x: float
        :param gamma_in: inner slope
        :type gamma_in: float
        :return: integrand of the deflection angel integral
        """
        return (y + x) ** (gamma_in - 4) * (1 - np.sqrt(1 - y**2))

    def alpha(self, R, Rs, alpha_Rs, gamma_in):
        """Deflection angel of gNFW profile along the radial direction.

        :param R: radius of interest
        :type R: float/numpy array
        :param Rs: scale radius
        :type Rs: float
        :param alpha_Rs: deflection at Rs
        :type alpha_Rs: float
        :param gamma_in: inner slope
        :type gamma_in: float
        :return: deflection angel at radius R
        :rtype: float
        """
        kappa_s = self.alpha_Rs_to_kappa_s(Rs, alpha_Rs, gamma_in)
        return self._alpha(R, Rs, kappa_s, gamma_in)

    def _alpha(self, R, Rs, kappa_s, gamma_in):
        """Deflection angel of gNFW profile along the radial direction.

        :param R: radius of interest
        :type R: float/numpy array
        :param Rs: scale radius
        :type Rs: float
        :param kappa_s: convergence at `Rs`
        :type kappa_s: float
        :param gamma_in: inner slope
        :type gamma_in: float
        :return: deflection angel at radius R
        :rtype: float
        """
        # R = np.maximum(R, self._s)
        x = R / Rs
        x = np.maximum(x, self._s)

        integral = self._integrate(self._alpha_integrand, x, gamma_in)

        alpha = (
            4
            * kappa_s
            * Rs
            * x ** (2 - gamma_in)
            * (
                hyp2f1(3 - gamma_in, 3 - gamma_in, 4 - gamma_in, -x) / (3 - gamma_in)
                + integral
            )
        )

        return alpha

    def kappa(self, R, Rs, alpha_Rs, gamma_in):
        """Convergence of gNFW profile along the radial direction.

        :param R: radius of interest
        :type R: float/numpy array
        :param Rs: scale radius
        :type Rs: float
        :param alpha_Rs: deflection at Rs
        :type alpha_Rs: float
        :param gamma_in: inner slope
        :type gamma_in: float
        :return: convergence at radius R
        :rtype: float
        """
        kappa_s = self.alpha_Rs_to_kappa_s(Rs, alpha_Rs, gamma_in)
        return self._kappa(R, Rs, kappa_s, gamma_in)

    def _kappa(self, R, Rs, kappa_s, gamma_in):
        """Convergence of gNFW profile along the radial direction.

        :param R: radius of interest
        :type R: float/numpy array
        :param Rs: scale radius
        :type Rs: float
        :param kappa_s: convergence at `Rs`
        :type kappa_s: float
        :param gamma_in: inner slope
        :type gamma_in: float
        :return: convergence at radius R
        :rtype: float
        """
        x = R / Rs
        x = np.maximum(x, self._s)

        integral = self._integrate(self._kappa_integrand, x, gamma_in)

        kappa = (
            2
            * kappa_s
            * x ** (1 - gamma_in)
            * ((1 + x) ** (gamma_in - 3) + (3 - gamma_in) * integral)
        )

        return kappa

    def kappa_s_to_alpha_Rs(self, kappa_s, Rs, gamma_in):
        """Convert the convergence at Rs to the density normalization.

        :param kappa_s: convergence at `Rs`
        :type kappa_s: float
        :param Rs: scale radius
        :type Rs: float
        :param gamma_in: inner slope
        :type gamma_in: float
        :return: rho0
        :rtype: float
        """
        return self._alpha(R=Rs, Rs=Rs, kappa_s=kappa_s, gamma_in=gamma_in)

    def alpha_Rs_to_kappa_s(self, Rs, alpha_Rs, gamma_in):
        """Convert the deflection at Rs to the convergence at Rs.

        :param Rs: scale radius
        :type Rs: float
        :param alpha_Rs: deflection at Rs
        :type alpha_Rs: float
        :param gamma_in: inner slope
        :type gamma_in: float
        :return: kappa_s
        :rtype: float
        """
        alpha_Rs_for_kappa_s_1 = self.get_alpha_Rs_for_kappa_s_1(Rs, gamma_in)
        kappa_s = alpha_Rs / alpha_Rs_for_kappa_s_1

        return kappa_s

    def rho02alpha(self, rho0, Rs, gamma_in):
        """Convenience function to compute alpha_Rs from rho0.

        :param rho0: density normalization (characteristic density)
        :type rho0: float
        :param Rs: scale radius
        :type Rs: float
        :param gamma_in: inner slope
        :type gamma_in: float
        :return: alpha_Rs
        :rtype: float
        """
        kappa_s = self.rho0_to_kappa_s(rho0, Rs)
        return self.kappa_s_to_alpha_Rs(kappa_s, Rs, gamma_in)

    def alpha2rho0(self, alpha_Rs, Rs, gamma_in):
        """Convenience function to compute rho0 from alpha_Rs.

        :param alpha_Rs: deflection at Rs
        :type alpha_Rs: float
        :param Rs: scale radius
        :type Rs: float
        :param gamma_in: inner slope
        :type gamma_in: float
        :return: rho0
        :rtype: float
        """
        kappa_s = self.alpha_Rs_to_kappa_s(Rs, alpha_Rs, gamma_in)
        return self.kappa_s_to_rho0(kappa_s, Rs)

    @staticmethod
    def rho0_to_kappa_s(rho0, Rs):
        """Convenience function to compute rho0 from alpha_Rs.

        :param rho0: density normalization (characteristic density)
        :type rho0: float
        :param Rs: scale radius
        :type Rs: float
        :param gamma_in: inner slope
        :type gamma_in: float
        :return: kappa_s
        :rtype: float
        """
        return rho0 * Rs

    @staticmethod
    def kappa_s_to_rho0(kappa_s, Rs):
        """Convenience function to compute rho0 from kappa_s. The returned rho_0 is
        normalized with $\\Sigma_{\\rm crit}$.

        :param kappa_s: convergence at `Rs`
        :type kappa_s: float
        :param Rs: scale radius
        :type Rs: float
        :param gamma_in: inner slope
        :type gamma_in: float
        :return: rho0
        :rtype: float
        """
        return kappa_s / Rs

    def get_alpha_Rs_for_kappa_s_1(self, Rs, gamma_in):
        """Compute the deflection at Rs.

        :param Rs: scale radius
        :type Rs: float
        :param gamma_in: inner slope
        :type gamma_in: float
        :return: alpha_Rs for kappa_s = 1
        :rtype: float
        """
        return Rs * self._alpha_1_interp(gamma_in)



__author__ = "dgilman"

from scipy.optimize import minimize
import numpy as np
from lenstronomy.Sampling.Samplers.pso import ParticleSwarmOptimizer
from lenstronomy.LensModel.QuadOptimizer.multi_plane_fast import MultiplaneFast
from lenstronomy.Sampling.Pool.pool import choose_pool
from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.LensModel.Util.decouple_multi_plane_util import (
    coordinates_and_deflections,
    decoupled_multiplane_class_setup,
    setup_lens_model,
)

__all__ = ["Optimizer"]


class Optimizer(object):
    """Class which executes the optimization routines. Currently implemented as a
    particle swarm optimization followed by a downhill simplex routine.

    Particle swarm optimizer is modified from the CosmoHammer particle swarm routine
    with different convergence criteria implemented.
    """

    def __init__(
        self,
        x_image,
        y_image,
        ray_shooting_class,
        parameter_class,
        tol_source=1e-5,
        tol_simplex_func=1e-3,
        simplex_n_iterations=400,
        pso_convergence_mean=50000,
        particle_swarm=True,
        re_optimize=False,
        re_optimize_scale=1.0,
        kwargs_multiplane_model=None,
    ):
        """

        :param x_image: x-coordinate of image positions
        :param y_image: y-coordinate of image positions
        :param ray_shooting_class: a class with a "ray_shooting" or "ray_shooting_fast" method; for example, an instance of LensModel, or of MultiPlaneFast
        :param parameter_class: a class that handles the lens model parameters being solved for, see classes in param_manager for examples
        :param tol_source: tolerance in the source plane that acts to punish poor solutions of the lens equation
        :param tol_simplex_func: the tolerence for the downhill simplex optimization routine to terminate
        :param simplex_n_iterations: the maximum number of iterations to iterate the downhill simplex optimization
        :param pso_convergence_mean: implements an early termination of the particle swarm optimization when the avergae penalty function hits this value
        :param particle_swarm: bool; if True, performs a particle swarm optimization of the lens model parameters before the downhill simplex routine
        :param re_optimize: bool; if True, initializes the particle swarm cloud in a narrow volume around the initial position
        :param re_optimize_scale: scales the size of the particle swarm cloud if re_optimize is True
        :param kwargs_multiplane_model: keyword arguments passed to MultiPlaneDecoupled if one is using the decoupled multi-plane formalism
        """
        self.x_image = x_image
        self.y_image = y_image
        self.ray_shooting_class = ray_shooting_class
        if callable(getattr(self.ray_shooting_class, "ray_shooting_fast", None)):
            self.ray_shooting_method = self.ray_shooting_class.ray_shooting_fast
        else:
            self.ray_shooting_method = self.ray_shooting_class.ray_shooting
        self._tol_source = tol_source
        self._pso_convergence_mean = pso_convergence_mean
        self._param_class = parameter_class
        self._tol_simplex_func = tol_simplex_func
        self._simplex_n_iterations = simplex_n_iterations
        self._particle_swarm = particle_swarm
        self._re_optimize = re_optimize
        self._re_optimize_scale = re_optimize_scale
        self._kwargs_multiplane_model = kwargs_multiplane_model

    @classmethod
    def decoupled_multiplane(
        cls,
        x_image,
        y_image,
        lens_model,
        kwargs_lens_model,
        index_lens_split,
        parameter_class,
        particle_swarm=True,
        re_optimize=False,
        re_optimize_scale=1.0,
        pso_convergence_mean=50000,
        tol_source=1e-5,
        tol_simplex_func=1e-3,
        simplex_n_iterations=400,
    ):
        """Initializes the Optimizer class using the decoupled multi-plane formalism for
        the lens model and ray shooting methods.

        :param x_image: x_image to fit (should be length 4)
        :param y_image: y_image to fit (should be length 4)
        :param lens_model: an instance of LensModel; should contain all deflectors along the line of sight
        :param kwargs_lens_model: keyword arguments for the LensModel class
        :param index_lens_split: a list of indexes where one splits the lens model, see documentation in Util/decoupled_multiplane_util/decoupled_multiplane_class_setup
        :param parameter_class: a class that handles the lens model parameters being solved for, see classes in param_manager for examples
        :param particle_swarm: bool; if True, performs a particle swarm optimization of the lens model parameters before the downhill simplex routine
        :param re_optimize: bool; if True, initializes the particle swarm cloud in a narrow volume around the initial position
        :param re_optimize_scale: scales the size of the particle swarm cloud if re_optimize is True
        :param pso_convergence_mean: implements an early termination of the particle swarm optimization when the avergae penalty function hits this value
        :param tol_source: tolerance in the source plane that acts to punish poor solutions of the lens equation
        :param tol_simplex_func: the tolerence for the downhill simplex optimization routine to terminate
        :param simplex_n_iterations: the maximum number of iterations to iterate the downhill simplex optimization
        :return: an instance of the Optimizer class using the decoupled multi-plane
         formalism for ray tracing.
        """
        (
            lens_model_fixed,
            lens_model_free,
            kwargs_lens_fixed,
            kwargs_lens_free,
            z_source,
            z_split,
            cosmo_bkg,
        ) = setup_lens_model(lens_model, kwargs_lens_model, index_lens_split)
        (
            x,
            y,
            alpha_x_foreground,
            alpha_y_foreground,
            alpha_beta_subx,
            alpha_beta_suby,
        ) = coordinates_and_deflections(
            lens_model_fixed,
            lens_model_free,
            kwargs_lens_fixed,
            kwargs_lens_free,
            x_image,
            y_image,
            z_split,
            z_source,
            cosmo_bkg,
        )
        npix, interp_points = None, None
        coordinate_type = "MULTIPLE_IMAGES"
        kwargs_decoupled = decoupled_multiplane_class_setup(
            lens_model_free,
            x,
            y,
            alpha_x_foreground,
            alpha_y_foreground,
            alpha_beta_subx,
            alpha_beta_suby,
            z_split,
            coordinate_type,
            interp_points,
            x_image,
            y_image,
        )
        lens_model_decoupled = LensModel(**kwargs_decoupled)
        # we have to reset the keyword arguments of the parameter class here
        parameter_class.kwargs_lens = kwargs_lens_free

        return Optimizer(
            x_image,
            y_image,
            lens_model_decoupled,
            parameter_class,
            tol_source,
            tol_simplex_func,
            simplex_n_iterations,
            pso_convergence_mean,
            particle_swarm,
            re_optimize,
            re_optimize_scale,
            kwargs_decoupled["kwargs_multiplane_model"],
        )

    @classmethod
    def full_raytracing(
        cls,
        x_image,
        y_image,
        lens_model_list,
        redshift_list,
        z_lens,
        z_source,
        parameter_class,
        astropy_instance=None,
        numerical_alpha_class=None,
        particle_swarm=True,
        re_optimize=False,
        re_optimize_scale=1.0,
        pso_convergence_mean=50000,
        foreground_rays=None,
        tol_source=1e-5,
        tol_simplex_func=1e-3,
        simplex_n_iterations=400,
    ):
        """

        :param x_image: x_image to fit (should be length 4)
        :param y_image: y_image to fit (should be length 4)
        :param lens_model_list: list of lens models for the system
        :param redshift_list: list of lens redshifts for the system
        :param z_lens: the main deflector redshift, the lens models being optimizer must be at this redshift
        :param z_source: the source redshift
        :param parameter_class: an instance of ParamClass (see documentation in QuadOptimizer.param_manager)
        :param astropy_instance: an instance of astropy to pass to the lens model
        :param numerical_alpha_class: a class to compute numerical deflection angles to pass to the lens model
        :param particle_swarm: bool, whether or not to use a PSO fit first
        :param re_optimize: bool, if True the initial spread of particles will be very tight
        :param re_optimize_scale: float, controls how tight the initial spread of particles is
        :param pso_convergence_mean: when to terminate the PSO fit
        :param foreground_rays: (optional) can pass in pre-computed foreground light rays from a previous fit
         so as to not waste time recomputing them
        :param tol_source: sigma in the source plane chi^2
        :param tol_simplex_func: tolerance for the downhill simplex optimization
        :param simplex_n_iterations: number of iterations per dimension for the downhill simplex optimization
        """

        fast_rayshooting = MultiplaneFast(
            x_image,
            y_image,
            z_lens,
            z_source,
            lens_model_list,
            redshift_list,
            astropy_instance,
            parameter_class,
            foreground_rays,
            tol_source,
            numerical_alpha_class,
        )
        return Optimizer(
            x_image,
            y_image,
            fast_rayshooting,
            parameter_class,
            tol_source,
            tol_simplex_func,
            simplex_n_iterations,
            pso_convergence_mean,
            particle_swarm,
            re_optimize,
            re_optimize_scale,
        )

    @property
    def kwargs_multiplane_model(self):
        """

        :return: the keyword arguments for the decoupled multi-plane class if they are specified
        """
        return self._kwargs_multiplane_model

    def optimize(
        self, n_particles=50, n_iterations=250, verbose=False, threadCount=1, seed=None
    ):
        """

        :param n_particles: number of PSO particles, will be ignored if self._particle_swarm is False
        :param n_iterations: number of PSO iterations, will be ignored if self._particle_swarm is False
        :param verbose: whether to print stuff
        :param threadCount: integer; number of threads in multi-threading mode
        :param seed: sets a random seed for reproducibility
        :return: keyword arguments that map (x_image, y_image) to the same source coordinate (source_x, source_y)
        """
        if seed is not None:
            np.random.seed(seed)
        if self._particle_swarm:
            if threadCount > 1:
                pool = choose_pool(mpi=False, processes=threadCount)
            else:
                pool = None
            kwargs = self._fit_pso(n_particles, n_iterations, pool, verbose)

        else:
            kwargs = self._param_class.kwargs_lens
        kwargs_lens_final, source_penalty = self._fit_amoeba(kwargs, verbose)
        source_x_array, source_y_array = self.ray_shooting_method(
            self.x_image, self.y_image, kwargs
        )
        source_x, source_y = np.mean(source_x_array), np.mean(source_y_array)

        if verbose:
            print("optimization done.")
            print("Recovered source position: ", (source_x_array, source_y_array))

        return kwargs_lens_final, [source_x, source_y]

    def _fit_pso(self, n_particles, n_iterations, pool, verbose):
        """Executes the PSO."""

        low_bounds, high_bounds = self._param_class.bounds(
            self._re_optimize, self._re_optimize_scale
        )

        pso = ParticleSwarmOptimizer(
            self._logL,
            low_bounds,
            high_bounds,
            n_particles,
            pool,
            args=[self._tol_source],
        )

        args_best, info = pso.optimize(
            n_iterations, verbose, early_stop_tolerance=self._pso_convergence_mean
        )
        kwargs = self._param_class.args_to_kwargs(args_best)
        if verbose:
            print("PSO done... ")
            print(
                "source plane chi^2: ",
                self.source_plane_penalty(args_best),
            )
            print("total chi^2: ", self._penalty_function(args_best))
        return kwargs

    def _fit_amoeba(self, kwargs, verbose):
        """Executes the downhill simplex."""

        args_init = self._param_class.kwargs_to_args(kwargs)
        options = {
            "adaptive": True,
            "fatol": self._tol_simplex_func,
            "maxiter": self._simplex_n_iterations * len(args_init),
        }
        method = "Nelder-Mead"
        if verbose:
            print("starting amoeba... ")
        opt = minimize(
            self._penalty_function,
            x0=args_init,
            method=method,
            options=options,
        )
        kwargs = self._param_class.args_to_kwargs(opt["x"])
        source_penalty = opt["fun"]
        return kwargs, source_penalty

    def _logL(self, args_lens, *args, **kwargs):
        """

        :param args: array of lens model parameters being optimized, computed from kwargs_lens in a specified
         param_class, see documentation in QuadOptimizer.param_manager
        :return: the log likelihood corresponding to the given chi^2
        """
        chi_square = self._penalty_function(args_lens)
        return -0.5 * chi_square

    def _penalty_function(self, args_lens, *args, **kwargs):
        """This function evaluates a metric that determines goodness of fit.

        :param args_lens: array of parameters that will be turned into keyword arguments
        :return: log-likelihood.
        """
        source_plane_chi2 = self.source_plane_penalty(args_lens)
        param_penalty = self._param_class.param_chi_square_penalty(args_lens)
        return source_plane_chi2 + param_penalty

    def source_plane_penalty(self, args_lens):
        """

        :param args_lens: array of lens model parameters being optimized, computed from kwargs_lens in a specified
         param_class, see documentation in QuadOptimizer.param_manager
        :return: chi2 penalty for the source position (all images must map to the same source coordinate)
        """
        kwargs_lens = self._param_class.args_to_kwargs(args_lens)
        betax, betay = self.ray_shooting_method(self.x_image, self.y_image, kwargs_lens)
        dx_source = (
            (betax[0] - betax[1]) ** 2
            + (betax[0] - betax[2]) ** 2
            + (betax[0] - betax[3]) ** 2
            + (betax[1] - betax[2]) ** 2
            + (betax[1] - betax[3]) ** 2
            + (betax[2] - betax[3]) ** 2
        )
        dy_source = (
            (betay[0] - betay[1]) ** 2
            + (betay[0] - betay[2]) ** 2
            + (betay[0] - betay[3]) ** 2
            + (betay[1] - betay[2]) ** 2
            + (betay[1] - betay[3]) ** 2
            + (betay[2] - betay[3]) ** 2
        )
        chi_square = 0.5 * (dx_source + dy_source) / self._tol_source**2
        return chi_square



"""Provisional Euclid instrument and observational settings.

See Optics and Observation Conditions spreadsheet at
https://docs.google.com/spreadsheets/d/1pMUB_OOZWwXON2dd5oP8PekhCT5MBBZJO1HV7IMZg4Y/edit?usp=sharing
for list of
sources.
"""

import lenstronomy.Util.util as util

__all__ = ["Euclid"]

VIS_obs = {
    "exposure_time": 565.0,
    "sky_brightness": 22.35,
    "magnitude_zero_point": 24.0,
    "num_exposures": 4,
    "seeing": 0.16,
    "psf_type": "GAUSSIAN",
}

# - keyword exposure_time: exposure time per image (in seconds)
# - keyword sky_brightness: sky brightness (in magnitude per square arcseconds in units of electrons)
# - keyword magnitude_zero_point: magnitude in which 1 count (e-) per second per arcsecond square is registered
# - keyword num_exposures: number of exposures that are combined (depends on coadd_years)
# - keyword seeing: Full-Width-at-Half-Maximum (FWHM) of PSF
# - keyword psf_type: string, type of PSF ('GAUSSIAN' supported)


class Euclid(object):
    """Class contains Euclid instrument and observation configurations."""

    def __init__(self, band="VIS", psf_type="GAUSSIAN", coadd_years=6):
        """

        :param band: string, only 'VIS' supported. Determines obs dictionary.
        :param psf_type: string, type of PSF ('GAUSSIAN' supported).
        :param coadd_years: int, number of years corresponding to num_exposures in obs dict. Currently supported: 2-6.
        """
        self.obs = VIS_obs
        if band != "VIS":
            raise ValueError("band %s not supported! Choose 'VIS'." % band)

        if psf_type != "GAUSSIAN":
            raise ValueError("psf_type %s not supported!" % psf_type)

        if coadd_years > 6 or coadd_years < 2:
            raise ValueError(
                " %s coadd_years not supported! Choose an integer between 2 and 6."
                % coadd_years
            )
        elif coadd_years != 6:
            self.obs["num_exposures"] = (coadd_years * VIS_obs["num_exposures"]) // 6

        self.camera = {"read_noise": 4.2, "pixel_scale": 0.101, "ccd_gain": 3.1}

        # - keyword read_noise: std of noise generated by read-out (in units of electrons)
        # - keyword pixel_scale: scale (in arcseconds) of pixels
        # - keyword ccd_gain: electrons/ADU (analog-to-digital unit)

    def kwargs_single_band(self):
        """

        :return: merged kwargs from camera and obs dicts
        """
        kwargs = util.merge_dicts(self.camera, self.obs)
        return kwargs



__author__ = "sibirrer"

import lenstronomy.Util.param_util as param_util
from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase

__all__ = ["Convergence"]


class Convergence(LensProfileBase):
    """A single mass sheet (external convergence)"""

    model_name = "CONVERGENCE"
    param_names = ["kappa", "ra_0", "dec_0"]
    lower_limit_default = {"kappa": -10, "ra_0": -100, "dec_0": -100}
    upper_limit_default = {"kappa": 10, "ra_0": 100, "dec_0": 100}

    def function(self, x, y, kappa, ra_0=0, dec_0=0):
        """Lensing potential.

        :param x: x-coordinate
        :param y: y-coordinate
        :param kappa: (external) convergence
        :return: lensing potential
        """
        theta, phi = param_util.cart2polar(x - ra_0, y - dec_0)
        f_ = 1.0 / 2 * kappa * theta**2
        return f_

    def derivatives(self, x, y, kappa, ra_0=0, dec_0=0):
        """Deflection angle.

        :param x: x-coordinate
        :param y: y-coordinate
        :param kappa: (external) convergence
        :return: deflection angles (first order derivatives)
        """
        x_ = x - ra_0
        y_ = y - dec_0
        f_x = kappa * x_
        f_y = kappa * y_
        return f_x, f_y

    def hessian(self, x, y, kappa, ra_0=0, dec_0=0):
        """Hessian matrix.

        :param x: x-coordinate
        :param y: y-coordinate
        :param kappa: external convergence
        :param ra_0: zero point of polynomial expansion (no deflection added)
        :param dec_0: zero point of polynomial expansion (no deflection added)
        :return: second order derivatives f_xx, f_xy, f_yx, f_yy
        """
        gamma1 = 0
        gamma2 = 0
        kappa = kappa
        f_xx = kappa + gamma1
        f_yy = kappa - gamma1
        f_xy = gamma2
        return f_xx, f_xy, f_xy, f_yy



__author__ = "dangilman"

import numpy as np
import pytest
import numpy.testing as npt
import lenstronomy.Util.param_util as param_util
from lenstronomy.Util import util
from lenstronomy.LensModel.Profiles.epl import EPL
from lenstronomy.LensModel.Profiles.multipole import Multipole, EllipticalMultipole
from lenstronomy.LensModel.Profiles.epl_multipole_m3m4 import (
    EPL_MULTIPOLE_M3M4,
    EPL_MULTIPOLE_M3M4_ELL,
)
from lenstronomy.LensModel.lens_model import LensModel


class TestEPL_MULTIPOLE_M3M4(object):
    """Test TestEPL_MULTIPOLE_M3M4 vs EPL + 2 MULTIPOLE (Spherical) values."""

    def setup_method(self):
        self.epl = EPL()
        self.multipole = Multipole()
        self.epl_multipole = EPL_MULTIPOLE_M3M4()
        self.x, self.y = util.make_grid(numPix=10, deltapix=0.2)
        self.lens_model_split = LensModel(["EPL", "MULTIPOLE", "MULTIPOLE"])
        self.lens_model = LensModel(["EPL_MULTIPOLE_M3M4"])
        self.lens_model_eplboxydisky = LensModel(["EPL_BOXYDISKY"])

    def test_function(self):
        a3_a = 0.008
        a4_a = -0.01
        delta_phi_m3 = np.pi / 8
        delta_phi_m4 = -np.pi / 6
        kwargs_epl = {
            "theta_E": 1.2,
            "center_x": 0.1,
            "center_y": 0.0,
            "e1": 0.1,
            "e2": 0.2,
            "gamma": 2.0,
        }
        kwargs_epl_multipole_m3m4 = {
            "theta_E": 1.2,
            "center_x": 0.1,
            "center_y": 0.0,
            "e1": 0.1,
            "e2": 0.2,
            "gamma": 2.0,
        }
        kwargs_epl_multipole_m3m4["a4_a"] = a4_a
        kwargs_epl_multipole_m3m4["a3_a"] = a3_a
        kwargs_epl_multipole_m3m4["delta_phi_m3"] = delta_phi_m3
        kwargs_epl_multipole_m3m4["delta_phi_m4"] = delta_phi_m4
        kwargs_epl_multipole_m3m4 = [kwargs_epl_multipole_m3m4]

        phi_q, q = param_util.ellipticity2phi_q(
            kwargs_epl_multipole_m3m4[0]["e1"], kwargs_epl_multipole_m3m4[0]["e2"]
        )
        rescale = kwargs_epl_multipole_m3m4[0]["theta_E"] / np.sqrt(q)
        kwargs_multliple_m3 = {
            "m": 3,
            "center_x": 0.1,
            "center_y": 0.0,
            "a_m": a3_a * rescale,
            "phi_m": phi_q + delta_phi_m3,
        }
        kwargs_multliple_m4 = {
            "m": 4,
            "center_x": 0.1,
            "center_y": 0.0,
            "a_m": a4_a * rescale,
            "phi_m": phi_q + delta_phi_m4,
        }
        kwargs_split = [kwargs_epl, kwargs_multliple_m3, kwargs_multliple_m4]
        function_joint = self.lens_model.potential(
            self.x, self.y, kwargs_epl_multipole_m3m4
        )
        function_split = self.lens_model_split.potential(self.x, self.y, kwargs_split)
        npt.assert_allclose(function_joint, function_split)

        kwargs_epl_boxydisky = {
            "theta_E": 1.2,
            "center_x": 0.1,
            "center_y": 0.0,
            "e1": 0.1,
            "e2": 0.2,
            "gamma": 2.0,
            "a4_a": a4_a,
        }
        kwargs_epl_multipole_m3m4[0]["a3_a"] = 0.0
        kwargs_epl_multipole_m3m4[0]["delta_phi_m4"] = 0.0
        function_joint = self.lens_model.potential(
            self.x, self.y, kwargs_epl_multipole_m3m4
        )
        function_eplboxy_disky = self.lens_model_eplboxydisky.potential(
            self.x, self.y, [kwargs_epl_boxydisky]
        )
        npt.assert_allclose(function_joint, function_eplboxy_disky)

    def test_derivatives(self):
        a3_a = 0.008
        a4_a = -0.01
        delta_phi_m3 = np.pi / 8
        delta_phi_m4 = -np.pi / 6
        kwargs_epl = {
            "theta_E": 1.2,
            "center_x": 0.1,
            "center_y": 0.0,
            "e1": 0.1,
            "e2": 0.2,
            "gamma": 2.0,
        }
        kwargs_epl_multipole_m3m4 = {
            "theta_E": 1.2,
            "center_x": 0.1,
            "center_y": 0.0,
            "e1": 0.1,
            "e2": 0.2,
            "gamma": 2.0,
        }
        kwargs_epl_multipole_m3m4["a4_a"] = a4_a
        kwargs_epl_multipole_m3m4["a3_a"] = a3_a
        kwargs_epl_multipole_m3m4["delta_phi_m3"] = delta_phi_m3
        kwargs_epl_multipole_m3m4["delta_phi_m4"] = delta_phi_m4
        kwargs_epl_multipole_m3m4 = [kwargs_epl_multipole_m3m4]

        phi_q, q = param_util.ellipticity2phi_q(
            kwargs_epl_multipole_m3m4[0]["e1"], kwargs_epl_multipole_m3m4[0]["e2"]
        )
        rescale = kwargs_epl_multipole_m3m4[0]["theta_E"] / np.sqrt(q)
        kwargs_multliple_m3 = {
            "m": 3,
            "center_x": 0.1,
            "center_y": 0.0,
            "a_m": a3_a * rescale,
            "phi_m": phi_q + delta_phi_m3,
        }
        kwargs_multliple_m4 = {
            "m": 4,
            "center_x": 0.1,
            "center_y": 0.0,
            "a_m": a4_a * rescale,
            "phi_m": phi_q + delta_phi_m4,
        }
        kwargs_split = [kwargs_epl, kwargs_multliple_m3, kwargs_multliple_m4]
        alpha_x, alpha_y = self.lens_model.alpha(
            self.x, self.y, kwargs_epl_multipole_m3m4
        )
        alpha_x_split, alpha_y_split = self.lens_model_split.alpha(
            self.x, self.y, kwargs_split
        )
        npt.assert_allclose(alpha_x, alpha_x_split)
        npt.assert_allclose(alpha_y, alpha_y_split)

        kwargs_epl_boxydisky = {
            "theta_E": 1.2,
            "center_x": 0.1,
            "center_y": 0.0,
            "e1": 0.1,
            "e2": 0.2,
            "gamma": 2.0,
            "a4_a": a4_a,
        }
        kwargs_epl_multipole_m3m4[0]["a3_a"] = 0.0
        kwargs_epl_multipole_m3m4[0]["delta_phi_m4"] = 0.0
        alpha_x_joint, alpha_y_joint = self.lens_model.alpha(
            self.x, self.y, kwargs_epl_multipole_m3m4
        )
        alpha_x_split, alpha_y_split = self.lens_model_eplboxydisky.alpha(
            self.x, self.y, [kwargs_epl_boxydisky]
        )
        npt.assert_allclose(alpha_x_joint, alpha_x_split)
        npt.assert_allclose(alpha_y_joint, alpha_y_split)

    def test_hessian(self):
        a3_a = 0.008
        a4_a = -0.01
        delta_phi_m3 = np.pi / 8
        delta_phi_m4 = -np.pi / 6
        kwargs_epl = {
            "theta_E": 1.2,
            "center_x": 0.1,
            "center_y": 0.0,
            "e1": 0.1,
            "e2": 0.2,
            "gamma": 2.0,
        }
        kwargs_epl_multipole_m3m4 = {
            "theta_E": 1.2,
            "center_x": 0.1,
            "center_y": 0.0,
            "e1": 0.1,
            "e2": 0.2,
            "gamma": 2.0,
        }
        kwargs_epl_multipole_m3m4["a4_a"] = a4_a
        kwargs_epl_multipole_m3m4["a3_a"] = a3_a
        kwargs_epl_multipole_m3m4["delta_phi_m3"] = delta_phi_m3
        kwargs_epl_multipole_m3m4["delta_phi_m4"] = delta_phi_m4
        kwargs_epl_multipole_m3m4 = [kwargs_epl_multipole_m3m4]

        phi_q, q = param_util.ellipticity2phi_q(
            kwargs_epl_multipole_m3m4[0]["e1"], kwargs_epl_multipole_m3m4[0]["e2"]
        )
        rescale = kwargs_epl_multipole_m3m4[0]["theta_E"] / np.sqrt(q)
        kwargs_multliple_m3 = {
            "m": 3,
            "center_x": 0.1,
            "center_y": 0.0,
            "a_m": a3_a * rescale,
            "phi_m": phi_q + delta_phi_m3,
        }
        kwargs_multliple_m4 = {
            "m": 4,
            "center_x": 0.1,
            "center_y": 0.0,
            "a_m": a4_a * rescale,
            "phi_m": phi_q + delta_phi_m4,
        }
        kwargs_split = [kwargs_epl, kwargs_multliple_m3, kwargs_multliple_m4]
        fxx, fxy, fyx, fyy = self.lens_model.hessian(
            self.x, self.y, kwargs_epl_multipole_m3m4
        )
        fxx_split, fxy_split, fyx_split, fyy_split = self.lens_model_split.hessian(
            self.x, self.y, kwargs_split
        )
        npt.assert_allclose(fxx, fxx_split)
        npt.assert_allclose(fxy, fxy_split)
        npt.assert_allclose(fyx, fyx_split)
        npt.assert_allclose(fyy, fyy_split)

        kwargs_epl_boxydisky = {
            "theta_E": 1.2,
            "center_x": 0.1,
            "center_y": 0.0,
            "e1": 0.1,
            "e2": 0.2,
            "gamma": 2.0,
            "a4_a": a4_a,
        }
        kwargs_epl_multipole_m3m4[0]["a3_a"] = 0.0
        kwargs_epl_multipole_m3m4[0]["delta_phi_m4"] = 0.0
        fxx, fxy, fyx, fyy = self.lens_model.hessian(
            self.x, self.y, kwargs_epl_multipole_m3m4
        )
        (
            fxx_split,
            fxy_split,
            fyx_split,
            fyy_split,
        ) = self.lens_model_eplboxydisky.hessian(self.x, self.y, [kwargs_epl_boxydisky])
        npt.assert_allclose(fxx, fxx_split)
        npt.assert_allclose(fxy, fxy_split)
        npt.assert_allclose(fyx, fyx_split)
        npt.assert_allclose(fyy, fyy_split)


class TestEPL_MULTIPOLE_M3M4_ELL(object):
    """Test TestEPL_MULTIPOLE_M3M4_ELL vs EPL + 2 MULTIPOLE_ELL values."""

    def setup_method(self):
        self.epl = EPL()
        self.multipole = EllipticalMultipole()
        self.epl_multipole = EPL_MULTIPOLE_M3M4_ELL()
        self.x, self.y = util.make_grid(numPix=10, deltapix=0.2)
        self.lens_model_split = LensModel(["EPL", "MULTIPOLE_ELL", "MULTIPOLE_ELL"])
        self.lens_model = LensModel(["EPL_MULTIPOLE_M3M4_ELL"])
        self.lens_model_eplboxydisky = LensModel(["EPL_BOXYDISKY_ELL"])

    def test_function(self):
        a3_a = 0.008
        a4_a = -0.01
        delta_phi_m3 = np.pi / 8
        delta_phi_m4 = -np.pi / 6
        kwargs_epl = {
            "theta_E": 1.2,
            "center_x": 0.1,
            "center_y": 0.0,
            "e1": 0.1,
            "e2": 0.2,
            "gamma": 2.0,
        }
        kwargs_epl_multipole_m3m4 = {
            "theta_E": 1.2,
            "center_x": 0.1,
            "center_y": 0.0,
            "e1": 0.1,
            "e2": 0.2,
            "gamma": 2.0,
        }
        kwargs_epl_multipole_m3m4["a4_a"] = a4_a
        kwargs_epl_multipole_m3m4["a3_a"] = a3_a
        kwargs_epl_multipole_m3m4["delta_phi_m3"] = delta_phi_m3
        kwargs_epl_multipole_m3m4["delta_phi_m4"] = delta_phi_m4
        kwargs_epl_multipole_m3m4 = [kwargs_epl_multipole_m3m4]

        phi_q, q = param_util.ellipticity2phi_q(
            kwargs_epl_multipole_m3m4[0]["e1"], kwargs_epl_multipole_m3m4[0]["e2"]
        )
        rescale = kwargs_epl_multipole_m3m4[0]["theta_E"] / np.sqrt(q)
        kwargs_multliple_m3 = {
            "m": 3,
            "center_x": 0.1,
            "center_y": 0.0,
            "a_m": a3_a * rescale,
            "phi_m": phi_q + delta_phi_m3,
            "q": q,
        }
        kwargs_multliple_m4 = {
            "m": 4,
            "center_x": 0.1,
            "center_y": 0.0,
            "a_m": a4_a * rescale,
            "phi_m": phi_q + delta_phi_m4,
            "q": q,
        }
        kwargs_split = [kwargs_epl, kwargs_multliple_m3, kwargs_multliple_m4]
        function_joint = self.lens_model.potential(
            self.x, self.y, kwargs_epl_multipole_m3m4
        )
        function_split = self.lens_model_split.potential(self.x, self.y, kwargs_split)
        npt.assert_allclose(function_joint, function_split)

        kwargs_epl_boxydisky = {
            "theta_E": 1.2,
            "center_x": 0.1,
            "center_y": 0.0,
            "e1": 0.1,
            "e2": 0.2,
            "gamma": 2.0,
            "a4_a": a4_a,
        }
        kwargs_epl_multipole_m3m4[0]["a3_a"] = 0.0
        kwargs_epl_multipole_m3m4[0]["delta_phi_m4"] = 0.0
        function_joint = self.lens_model.potential(
            self.x, self.y, kwargs_epl_multipole_m3m4
        )
        function_eplboxy_disky = self.lens_model_eplboxydisky.potential(
            self.x, self.y, [kwargs_epl_boxydisky]
        )
        npt.assert_allclose(function_joint, function_eplboxy_disky)

    def test_derivatives(self):
        a3_a = 0.008
        a4_a = -0.01
        delta_phi_m3 = np.pi / 8
        delta_phi_m4 = -np.pi / 6
        kwargs_epl = {
            "theta_E": 1.2,
            "center_x": 0.1,
            "center_y": 0.0,
            "e1": 0.1,
            "e2": 0.2,
            "gamma": 2.0,
        }
        kwargs_epl_multipole_m3m4 = {
            "theta_E": 1.2,
            "center_x": 0.1,
            "center_y": 0.0,
            "e1": 0.1,
            "e2": 0.2,
            "gamma": 2.0,
        }
        kwargs_epl_multipole_m3m4["a4_a"] = a4_a
        kwargs_epl_multipole_m3m4["a3_a"] = a3_a
        kwargs_epl_multipole_m3m4["delta_phi_m3"] = delta_phi_m3
        kwargs_epl_multipole_m3m4["delta_phi_m4"] = delta_phi_m4
        kwargs_epl_multipole_m3m4 = [kwargs_epl_multipole_m3m4]

        phi_q, q = param_util.ellipticity2phi_q(
            kwargs_epl_multipole_m3m4[0]["e1"], kwargs_epl_multipole_m3m4[0]["e2"]
        )
        rescale = kwargs_epl_multipole_m3m4[0]["theta_E"] / np.sqrt(q)
        kwargs_multliple_m3 = {
            "m": 3,
            "center_x": 0.1,
            "center_y": 0.0,
            "a_m": a3_a * rescale,
            "phi_m": phi_q + delta_phi_m3,
            "q": q,
        }
        kwargs_multliple_m4 = {
            "m": 4,
            "center_x": 0.1,
            "center_y": 0.0,
            "a_m": a4_a * rescale,
            "phi_m": phi_q + delta_phi_m4,
            "q": q,
        }
        kwargs_split = [kwargs_epl, kwargs_multliple_m3, kwargs_multliple_m4]
        alpha_x, alpha_y = self.lens_model.alpha(
            self.x, self.y, kwargs_epl_multipole_m3m4
        )
        alpha_x_split, alpha_y_split = self.lens_model_split.alpha(
            self.x, self.y, kwargs_split
        )
        npt.assert_allclose(alpha_x, alpha_x_split)
        npt.assert_allclose(alpha_y, alpha_y_split)

        kwargs_epl_boxydisky = {
            "theta_E": 1.2,
            "center_x": 0.1,
            "center_y": 0.0,
            "e1": 0.1,
            "e2": 0.2,
            "gamma": 2.0,
            "a4_a": a4_a,
        }
        kwargs_epl_multipole_m3m4[0]["a3_a"] = 0.0
        kwargs_epl_multipole_m3m4[0]["delta_phi_m4"] = 0.0
        alpha_x_joint, alpha_y_joint = self.lens_model.alpha(
            self.x, self.y, kwargs_epl_multipole_m3m4
        )
        alpha_x_split, alpha_y_split = self.lens_model_eplboxydisky.alpha(
            self.x, self.y, [kwargs_epl_boxydisky]
        )
        npt.assert_allclose(alpha_x_joint, alpha_x_split)
        npt.assert_allclose(alpha_y_joint, alpha_y_split)

    def test_hessian(self):
        a3_a = 0.008
        a4_a = -0.01
        delta_phi_m3 = np.pi / 8
        delta_phi_m4 = -np.pi / 6
        kwargs_epl = {
            "theta_E": 1.2,
            "center_x": 0.1,
            "center_y": 0.0,
            "e1": 0.1,
            "e2": 0.2,
            "gamma": 2.0,
        }
        kwargs_epl_multipole_m3m4 = {
            "theta_E": 1.2,
            "center_x": 0.1,
            "center_y": 0.0,
            "e1": 0.1,
            "e2": 0.2,
            "gamma": 2.0,
        }
        kwargs_epl_multipole_m3m4["a4_a"] = a4_a
        kwargs_epl_multipole_m3m4["a3_a"] = a3_a
        kwargs_epl_multipole_m3m4["delta_phi_m3"] = delta_phi_m3
        kwargs_epl_multipole_m3m4["delta_phi_m4"] = delta_phi_m4
        kwargs_epl_multipole_m3m4 = [kwargs_epl_multipole_m3m4]

        phi_q, q = param_util.ellipticity2phi_q(
            kwargs_epl_multipole_m3m4[0]["e1"], kwargs_epl_multipole_m3m4[0]["e2"]
        )
        rescale = kwargs_epl_multipole_m3m4[0]["theta_E"] / np.sqrt(q)
        kwargs_multliple_m3 = {
            "m": 3,
            "center_x": 0.1,
            "center_y": 0.0,
            "a_m": a3_a * rescale,
            "phi_m": phi_q + delta_phi_m3,
            "q": q,
        }
        kwargs_multliple_m4 = {
            "m": 4,
            "center_x": 0.1,
            "center_y": 0.0,
            "a_m": a4_a * rescale,
            "phi_m": phi_q + delta_phi_m4,
            "q": q,
        }
        kwargs_split = [kwargs_epl, kwargs_multliple_m3, kwargs_multliple_m4]
        fxx, fxy, fyx, fyy = self.lens_model.hessian(
            self.x, self.y, kwargs_epl_multipole_m3m4
        )
        fxx_split, fxy_split, fyx_split, fyy_split = self.lens_model_split.hessian(
            self.x, self.y, kwargs_split
        )
        npt.assert_allclose(fxx, fxx_split)
        npt.assert_allclose(fxy, fxy_split)
        npt.assert_allclose(fyx, fyx_split)
        npt.assert_allclose(fyy, fyy_split)

        kwargs_epl_boxydisky = {
            "theta_E": 1.2,
            "center_x": 0.1,
            "center_y": 0.0,
            "e1": 0.1,
            "e2": 0.2,
            "gamma": 2.0,
            "a4_a": a4_a,
        }
        kwargs_epl_multipole_m3m4[0]["a3_a"] = 0.0
        kwargs_epl_multipole_m3m4[0]["delta_phi_m4"] = 0.0
        fxx, fxy, fyx, fyy = self.lens_model.hessian(
            self.x, self.y, kwargs_epl_multipole_m3m4
        )
        (
            fxx_split,
            fxy_split,
            fyx_split,
            fyy_split,
        ) = self.lens_model_eplboxydisky.hessian(self.x, self.y, [kwargs_epl_boxydisky])
        npt.assert_allclose(fxx, fxx_split)
        npt.assert_allclose(fxy, fxy_split)
        npt.assert_allclose(fyx, fyx_split)
        npt.assert_allclose(fyy, fyy_split)


if __name__ == "__main__":
    pytest.main()



from lenstronomy.Util.magnification_finite_util import (
    auto_raytracing_grid_size,
    auto_raytracing_grid_resolution,
    setup_mag_finite,
)
from lenstronomy.LensModel.lens_model import LensModel
import numpy.testing as npt
from lenstronomy.LightModel.light_model import LightModel
import pytest


class TestMagnificationFiniteUtil(object):
    def test_grid_resolution(self):
        source_size = 1.0
        res_scale = 0.03
        ref = 15.0
        power = 1.2
        resolution = auto_raytracing_grid_resolution(source_size, res_scale, ref, power)

        npt.assert_almost_equal(resolution, res_scale * (source_size / ref) ** power)

    def test_grid_size(self):
        source_size = 20.0
        size = auto_raytracing_grid_size(source_size, grid_size_scale=0.04, power=0.9)

        npt.assert_almost_equal(size, 0.04 * source_size**0.9)

    def test_setup(self):
        grid_resolution = 0.001
        grid_radius_arcsec = 0.05
        source_light_model = ["GAUSSIAN"]
        source_model = LightModel(source_light_model)
        kwargs_source = [{"amp": 1, "sigma": 0.0408, "center_x": 0, "center_y": 0}]

        npt.assert_equal(True, grid_resolution is not None)
        npt.assert_equal(True, grid_radius_arcsec is not None)

        npt.assert_equal(True, grid_resolution == 0.001)
        npt.assert_equal(True, grid_radius_arcsec == 0.05)

        (
            gridx,
            gridy,
            source_model,
            kwargs_source,
            grid_resolution,
            grid_radius_arcsec,
        ) = setup_mag_finite(
            grid_radius_arcsec, grid_resolution, source_model, kwargs_source
        )


if __name__ == "__main__":
    pytest.main()



# Nautilus sampling
# https://github.com/johannesulf/nautilus

from lenstronomy.Sampling.Pool.pool import choose_pool
import time

__all__ = ["Nautilus"]


class Nautilus(object):
    def __init__(self, likelihood_module):
        """Importance nested sampling with Nautilus [1]_

        References:
        -----------
        .. [1] Johannes Lange, in prep, https://github.com/johannesulf/nautilus

        :param likelihood_module: LikelihoodModule() instance
        """
        self._likelihood_module = likelihood_module
        self._num_param, _ = self._likelihood_module.param.num_param()
        (
            self._lower_limit,
            self._upper_limit,
        ) = self._likelihood_module.param.param_limits()

    def nautilus_sampling(
        self,
        prior_type="uniform",
        mpi=False,
        thread_count=1,
        verbose=True,
        one_step=False,
        **kwargs_nautilus
    ):
        """

        :param prior_type: string; prior type. Currently only 'uniform' supported
         (in addition to Prior class in Likelihood module)
        :param mpi: MPI option (currently not supported)
        :param thread_count: integer; multi-threading option (currently not supported)
        :param verbose: verbose statements of Nautilus
        :param one_step: boolean, if True, only runs one iteration of filling the sampler and re-training.
         This is meant for test purposes of the sampler to operate with little computational effort
        :param kwargs_nautilus: additional keyword arguments for Nautilus
        :return: points, log_w, log_l, log_z
        """
        from nautilus import Prior, Sampler

        prior = Prior()
        # TODO better prior integration with Nautilus
        if prior_type == "uniform":
            for i in range(self._num_param):
                prior.add_parameter(dist=(self._lower_limit[i], self._upper_limit[i]))
        else:
            raise ValueError(
                "prior_type %s is not supported for Nautilus wrapper." % prior_type
            )
        # loop through prior
        pool = choose_pool(mpi=mpi, processes=thread_count, use_dill=True)
        sampler = Sampler(
            prior,
            likelihood=self.likelihood,
            pool=pool,
            pass_dict=False,
            **kwargs_nautilus
        )
        time_start = time.time()
        if one_step is True:
            sampler.add_bound()
            sampler.fill_bound()
        else:
            sampler.run(verbose=verbose)
        points, log_w, log_l = sampler.posterior(return_as_dict=False)
        log_z = sampler.evidence()
        time_end = time.time()
        if pool.is_master():
            print(time_end - time_start, "time taken for nested sampling")
        return points, log_w, log_l, log_z

    def likelihood(self, args):
        """Log likelihood.

        :param args: ctype
        :return: log likelihood
        """
        python_list = []
        for i in range(self._num_param):
            python_list.append(args[i])
        return self._likelihood_module.likelihood(a=python_list)



from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.LensModel.Profiles.radial_interpolated import RadialInterpolate
from lenstronomy.Util import util
import numpy as np
import numpy.testing as npt


class TestRadialInterpolation(object):
    """Testing radial interpolation class test case is a SIS profile and a constant mass
    sheet."""

    def setup_method(self):
        self.rad_interp = RadialInterpolate()

    def test_sis(self):
        interp_profile = LensModel(lens_model_list=["RADIAL_INTERPOL"])
        sis = LensModel(lens_model_list=["SIS"])
        kwargs_sis = [{"theta_E": 1, "center_x": 0, "center_y": 0}]
        r_bin_log = np.logspace(-4, 1, 200)
        kappa_r_sis = sis.kappa(r_bin_log, 0, kwargs=kwargs_sis)
        kwargs_interp = [{"r_bin": r_bin_log, "kappa_r": kappa_r_sis}]

        x, y = util.make_grid(numPix=10, deltapix=0.1)

        # hessian
        f_xx_int, f_xy_int, f_yx_int, f_yy_int = interp_profile.hessian(
            x, y, kwargs_interp
        )
        f_xx, f_xy, f_yx, f_yy = sis.hessian(x, y, kwargs_sis)
        kappa_int = 1 / 2 * (f_xx_int + f_yy_int)
        kappa = 1 / 2 * (f_xx + f_yy)

        gamma1 = 1 / 2 * (f_xx - f_yy)
        gamma1_int = 1 / 2 * (f_xx_int - f_yy_int)

        npt.assert_almost_equal(kappa_int / kappa, 1, decimal=3)
        npt.assert_almost_equal(f_xx_int, f_xx, decimal=2)
        npt.assert_almost_equal(f_xy_int, f_xy, decimal=2)
        npt.assert_almost_equal(f_yx_int, f_yx, decimal=2)
        npt.assert_almost_equal(f_yy_int, f_yy, decimal=2)

        # deflection
        f_x_int, f_y_int = interp_profile.alpha(x, y, kwargs_interp)
        f_x, f_y = sis.alpha(x, y, kwargs_sis)
        npt.assert_almost_equal(f_x_int, f_x, decimal=3)
        npt.assert_almost_equal(f_y_int, f_y, decimal=3)

        # potential
        f_int = interp_profile.potential(x, y, kwargs_interp)
        f_ = sis.potential(x, y, kwargs_sis)
        npt.assert_almost_equal(f_int, f_, decimal=3)

    def test_mass_sheet(self):
        interp_profile = LensModel(lens_model_list=["RADIAL_INTERPOL"])
        mass_sheet = LensModel(lens_model_list=["CONVERGENCE"])
        kwargs_convergence = [{"kappa": 0.5}]
        r_bin = np.linspace(start=0, stop=1, num=100)
        kappa_r_sis = mass_sheet.kappa(r_bin, 0, kwargs=kwargs_convergence)
        kwargs_interp = [{"r_bin": r_bin, "kappa_r": kappa_r_sis}]

        x, y = util.make_grid(numPix=10, deltapix=0.1)

        # hessian
        f_xx_int, f_xy_int, f_yx_int, f_yy_int = interp_profile.hessian(
            x, y, kwargs_interp
        )
        f_xx, f_xy, f_yx, f_yy = mass_sheet.hessian(x, y, kwargs_convergence)
        kappa_int = 1 / 2 * (f_xx_int + f_yy_int)
        kappa = 1 / 2 * (f_xx + f_yy)
        npt.assert_almost_equal(kappa_int / kappa, 1, decimal=3)

        npt.assert_almost_equal(f_xx_int, f_xx, decimal=3)
        npt.assert_almost_equal(f_xy_int, f_xy, decimal=3)
        npt.assert_almost_equal(f_yx_int, f_yx, decimal=3)
        npt.assert_almost_equal(f_yy_int, f_yy, decimal=3)

        # deflection
        f_x_int, f_y_int = interp_profile.alpha(x, y, kwargs_interp)
        f_x, f_y = mass_sheet.alpha(x, y, kwargs_convergence)
        npt.assert_almost_equal(f_x_int, f_x, decimal=3)
        npt.assert_almost_equal(f_y_int, f_y, decimal=3)

        # potential
        f_int = interp_profile.potential(x, y, kwargs_interp)
        f_ = mass_sheet.potential(x, y, kwargs_convergence)
        npt.assert_almost_equal(f_int, f_, decimal=3)



__author__ = "sibirrer"

import numpy as np
from scipy.interpolate import interp1d
import lenstronomy.GalKin.velocity_util as vel_util
from lenstronomy.GalKin.cosmo import Cosmo
from lenstronomy.GalKin.anisotropy import Anisotropy
from lenstronomy.LensModel.Profiles.spp import SPP
import lenstronomy.Util.constants as const
import math
from lenstronomy.GalKin.light_profile import LightProfile
from copy import deepcopy

__all__ = ["AnalyticKinematics"]


class AnalyticKinematics(Anisotropy):
    """Class to compute eqn 20 in Suyu+2010 with a Monte-Carlo from rendering from the
    light profile distribution and displacing them with a Gaussian seeing convolution.

    This class assumes spherical symmetry in light and mass distribution and
        - a Hernquist light profile (parameterised by the half-light radius)
        - a power-law mass profile (parameterized by the Einstein radius and logarithmic slop)

    The analytic equations for the kinematics in this approximation are presented e.g. in Suyu et al. 2010 and
    the spectral rendering approach to compute the seeing convolved slit measurement is presented in Birrer et al. 2016.
    The stellar anisotropy is parameterised based on Osipkov 1979; Merritt 1985.

    WARNING!!! Only supports Osipkov-Merritt anisotropy for now!

    All units are meant to be in angular arc seconds. The physical units are fold in through the angular diameter
    distances
    """

    def __init__(
        self,
        kwargs_cosmo,
        interpol_grid_num=200,
        log_integration=True,
        max_integrate=100,
        min_integrate=1e-4,
    ):
        """

        :param kwargs_cosmo: keyword argument with angular diameter distances entering the Galkin.cosmo class
        :param interpol_grid_num: number of interpolations in radius to compute radial velocity dispersion
        :param log_integration: perform numerical integration in logarithmic space,
            setting False may lead to less accurate results
        :param max_integrate: maximum radius of integration (in projected arc seconds)
        :param min_integrate: minimum drawing/calculation of velocity dispersion (in projected arc seconds)
        """
        self._interp_grid_num = interpol_grid_num
        self._log_int = log_integration
        self._max_integrate = (
            max_integrate  # maximal integration (and interpolation) in units of arcsecs
        )
        self._min_integrate = (
            min_integrate  # min integration (and interpolation) in units of arcsecs
        )
        self._max_interpolate = max_integrate  # we chose to set the interpolation range to the integration range
        self._min_interpolate = min_integrate  # we chose to set the interpolation range to the integration range

        self._cosmo = Cosmo(**kwargs_cosmo)
        self._spp = SPP()

        self.light_profile = LightProfile(
            ["HERNQUIST"],
            interpol_grid_num=interpol_grid_num,
            max_interpolate=max_integrate,
            min_interpolate=min_integrate,
            max_draw=max_integrate,
        )
        Anisotropy.__init__(self, anisotropy_type="OM")

    @property
    def max_integrate(self):
        """Get the maximum range of integration."""
        return self._max_integrate

    @property
    def min_integrate(self):
        """Get the minimum range of integration."""
        return self._min_integrate

    def _rho0_r0_gamma(self, theta_E, gamma):
        # equation (14) in Suyu+ 2010
        return (
            -1
            * math.gamma(gamma / 2)
            / (np.sqrt(np.pi) * math.gamma((gamma - 3) / 2.0))
            * theta_E**gamma
            / self._cosmo.arcsec2phys_lens(theta_E)
            * self._cosmo.epsilon_crit
            * const.M_sun
            / const.Mpc**3
        )

    @staticmethod
    def _get_hernquist_scale_radius(kwargs_light):
        """Returns the scale of the Hernquist light profile.

        :param kwargs_light: keyword arguments of the light profile
        """
        if "a" not in kwargs_light:
            if "Rs" in kwargs_light:
                a = kwargs_light["Rs"]
            elif "r_eff" in kwargs_light:
                a = 0.551 * kwargs_light["r_eff"]
            else:
                raise ValueError("Hernquist half-light radius can not be determined!")
        else:
            a = kwargs_light["a"]

        return a

    @classmethod
    def draw_light(cls, kwargs_light):
        """Draws a random light tracer particle from the Hernquist light profile.

        :param kwargs_light: keyword argument (list) of the light model
        :return: 3d radius (if possible), 2d projected radius, x-projected coordinate,
            y-projected coordinate
        """
        a = cls._get_hernquist_scale_radius(kwargs_light)

        r = vel_util.draw_hernquist(a)
        R, x, y = vel_util.project2d_random(r)
        return r, R, x, y

    def _sigma_s2(self, r, R, r_ani, a, gamma, rho0_r0_gamma):
        """Projected velocity dispersion :param r: 3d radius of the light tracer
        particle :param R: 2d projected radius of the light tracer particle :param
        r_ani: anisotropy radius :param a: scale of the Hernquist light profile :param
        gamma: power-law slope of the mass profile :param rho0_r0_gamma: combination of
        Einstein radius and power-law slope as equation (14) in Suyu+ 2010 :return:
        projected velocity dispersion."""
        beta = self.beta_r(r, **{"r_ani": r_ani})
        return (1 - beta * R**2 / r**2) * self._sigma_r2_interp(
            r, a, gamma, rho0_r0_gamma, r_ani
        )

    def sigma_s2(self, r, R, kwargs_mass, kwargs_light, kwargs_anisotropy):
        """Returns unweighted los velocity dispersion for a specified projected radius,
        with weight 1.

        :param r: 3d radius (not needed for this calculation)
        :param R: 2d projected radius (in angular units of arcsec)
        :param kwargs_mass: mass model parameters (following lenstronomy lens model
            conventions)
        :param kwargs_light: deflector light parameters (following lenstronomy light
            model conventions)
        :param kwargs_anisotropy: anisotropy parameters, may vary according to
            anisotropy type chosen. We refer to the Anisotropy() class for details on
            the parameters.
        :return: line-of-sight projected velocity dispersion at projected radius R from
            3d radius r
        """
        a, gamma, rho0_r0_gamma, r_ani = self._read_out_params(
            kwargs_mass, kwargs_light, kwargs_anisotropy
        )
        return self._sigma_s2(r, R, r_ani, a, gamma, rho0_r0_gamma), 1

    def sigma_r2(self, r, kwargs_mass, kwargs_light, kwargs_anisotropy):
        """Equation (19) in Suyu+ 2010.

        :param r: 3d radius
        :param kwargs_mass: mass profile keyword arguments
        :param kwargs_light: light profile keyword arguments
        :param kwargs_anisotropy: anisotropy keyword arguments
        :return: velocity dispersion in [m/s]
        """
        a, gamma, rho0_r0_gamma, r_ani = self._read_out_params(
            kwargs_mass, kwargs_light, kwargs_anisotropy
        )
        return self._sigma_r2(r, a, gamma, rho0_r0_gamma, r_ani)

    def _read_out_params(self, kwargs_mass, kwargs_light, kwargs_anisotropy):
        """Reads the relevant parameters out of the keyword arguments and transforms
        them to the conventions used in this class.

        :param kwargs_mass: mass profile keyword arguments
        :param kwargs_light: light profile keyword arguments
        :param kwargs_anisotropy: anisotropy keyword arguments
        :return: a (Rs of Hernquist profile), gamma, rho0_r0_gamma, r_ani
        """
        a = self._get_hernquist_scale_radius(kwargs_light)

        if "rho0_r0_gamma" not in kwargs_mass:
            kwargs_mass["rho0_r0_gamma"] = self._rho0_r0_gamma(
                kwargs_mass["theta_E"], kwargs_mass["gamma"]
            )

        gamma = kwargs_mass["gamma"]
        rho0_r0_gamma = kwargs_mass["rho0_r0_gamma"]
        r_ani = kwargs_anisotropy["r_ani"]

        return a, gamma, rho0_r0_gamma, r_ani

    def _sigma_r2(self, r, a, gamma, rho0_r0_gamma, r_ani):
        """Equation (19) in Suyu+ 2010."""
        # first term
        prefac1 = 4 * np.pi * const.G * a ** (-gamma) * rho0_r0_gamma / (3 - gamma)
        prefac2 = r * (r + a) ** 3 / (r**2 + r_ani**2)
        # TODO check whether interpolation functions can speed this up
        hyp1 = vel_util.hyp_2F1(a=2 + gamma, b=gamma, c=3 + gamma, z=1.0 / (1 + r / a))
        hyp2 = vel_util.hyp_2F1(a=3, b=gamma, c=1 + gamma, z=-a / r)
        fac = r_ani**2 / a**2 * hyp1 / (
            (2 + gamma) * (r / a + 1) ** (2 + gamma)
        ) + hyp2 / (gamma * (r / a) ** gamma)
        return (
            prefac1 * prefac2 * fac * (const.arcsec * self._cosmo.dd * const.Mpc) ** 2
        )

    def _sigma_r2_interp(self, r, a, gamma, rho0_r0_gamma, r_ani):
        """

        :param r:
        :param a:
        :param gamma:
        :param rho0_r0_gamma:
        :param r_ani:
        :return:
        """
        if not hasattr(self, "_interp_sigma_r2"):
            min_log = np.log10(self._min_integrate)
            max_log = np.log10(self._max_integrate)
            r_array = np.logspace(min_log, max_log, self._interp_grid_num)
            I_R_sigma2_array = []
            for r_i in r_array:
                I_R_sigma2_array.append(
                    self._sigma_r2(r_i, a, gamma, rho0_r0_gamma, r_ani)
                )
            self._interp_sigma_r2 = interp1d(
                np.log(r_array), np.array(I_R_sigma2_array), fill_value="extrapolate"
            )
        return self._interp_sigma_r2(np.log(r))

    def _I_R_sigma2(self, R, kwargs_mass, kwargs_light, kwargs_anisotropy):
        """Equation A15 in Mamon&Lokas 2005 as a logarithmic numerical integral (if
        option is chosen)

        :param R: 2d projected radius (in angular units)
        :param kwargs_mass: mass model parameters (following lenstronomy lens model
            conventions)
        :param kwargs_light: deflector light parameters (following lenstronomy light
            model conventions)
        :param kwargs_anisotropy: anisotropy parameters, may vary according to
            anisotropy type chosen. See the Anisotropy() class for details on the
            parameters.
        :return: integral of A15 in Mamon&Lokas 2005
        """
        R = max(R, self._min_integrate)
        max_integrate = (
            self._max_integrate
        )  # make sure the integration of the Jeans equation is performed further out than the interpolation

        if self._log_int is True:
            min_log = np.log(R)
            max_log = np.log(max_integrate)
            dlogr = (max_log - min_log) / (self._interp_grid_num - 1)
            r_array = np.logspace(
                min_log + dlogr / 2.0,
                max_log + dlogr / 2.0,
                self._interp_grid_num,
                base=np.e,
            )
            dlog_r = np.log(r_array[2]) - np.log(r_array[1])
            IR_sigma2_ = self._integrand_suyu10_eq21(
                r_array, R, kwargs_mass, kwargs_light, kwargs_anisotropy
            )
            IR_sigma2_dr = IR_sigma2_ * dlog_r * r_array
        else:
            r_array = np.linspace(
                start=R, stop=self._max_interpolate, num=self._interp_grid_num
            )
            dr = r_array[2] - r_array[1]
            IR_sigma2_ = self._integrand_suyu10_eq21(
                r_array + dr / 2.0, R, kwargs_mass, kwargs_light, kwargs_anisotropy
            )
            IR_sigma2_dr = IR_sigma2_ * dr

        IR_sigma2 = 2 * np.sum(IR_sigma2_dr)  # integral from angle to

        kwargs_light_copy = deepcopy(kwargs_light)
        if "r_eff" in kwargs_light_copy:
            kwargs_light_copy["Rs"] = kwargs_light_copy["r_eff"] * 0.551
            kwargs_light_copy["amp"] = 1
            kwargs_light_copy.pop("r_eff")
        IR = self.light_profile.light_2d_finite(R, [kwargs_light_copy])

        return IR_sigma2, IR

    def _integrand_suyu10_eq21(
        self, r, R, kwargs_mass, kwargs_light, kwargs_anisotropy
    ):
        """Compute the integrand of Eq. 21 from Suyu et al. (2010).

        :param r: 3d radius (not needed for this calculation)
        :param R: 2d projected radius (in angular units of arcsec)
        :param kwargs_mass: mass model parameters (following lenstronomy lens model
            conventions)
        :param kwargs_light: deflector light parameters (following lenstronomy light
            model conventions)
        :param kwargs_anisotropy: anisotropy parameters, may vary according to
            anisotropy type chosen. We refer to the Anisotropy() class for details on
            the parameters.
        :return: integrand of Eq. 21 from Suyu et al. (2010)
        """
        a, gamma, rho0_r0_gamma, r_ani = self._read_out_params(
            kwargs_mass, kwargs_light, kwargs_anisotropy
        )
        kwargs_light_copy = deepcopy(kwargs_light)
        if "r_eff" in kwargs_light_copy:
            kwargs_light_copy["Rs"] = kwargs_light_copy["r_eff"] * 0.551
            kwargs_light_copy["amp"] = 1
            kwargs_light_copy.pop("r_eff")

        return (
            self._sigma_s2(r, R, r_ani, a, gamma, rho0_r0_gamma)
            * r
            / np.sqrt(r**2 - R**2)
            * self.light_profile.light_3d(r, [kwargs_light_copy])
        )

    def I_R_sigma2_and_IR(self, R, kwargs_mass, kwargs_light, kwargs_anisotropy):
        """Return I(R)*sigma^2 equation 20 in Suyu 2010 as interpolation in log space,
        and I(R)

        :param R: projected radius
        :param kwargs_mass: mass profile keyword arguments
        :param kwargs_light: light model keyword arguments
        :param kwargs_anisotropy: stellar anisotropy keyword arguments
        :return:
        """
        return self._I_R_sigma2(R, kwargs_mass, kwargs_light, kwargs_anisotropy)

    def grav_potential(self, r, kwargs_mass):
        """Gravitational potential in SI units.

        :param r: radius (arc seconds)
        :param kwargs_mass:
        :return: gravitational potential
        """
        theta_E = kwargs_mass["theta_E"]
        gamma = kwargs_mass["gamma"]
        mass_dimless = self._spp.mass_3d_lens(r, theta_E, gamma)
        mass_dim = (
            mass_dimless
            * const.arcsec**2
            * self._cosmo.dd
            * self._cosmo.ds
            / self._cosmo.dds
            * const.Mpc
            * const.c**2
            / (4 * np.pi * const.G)
        )
        grav_pot = -const.G * mass_dim / (r * const.arcsec * self._cosmo.dd * const.Mpc)
        return grav_pot

    def delete_cache(self):
        """Deletes temporary cache tight to a specific model.

        :return:
        """
        if hasattr(self, "_interp_sigma_r2"):
            del self._interp_sigma_r2



from lenstronomy.LightModel.Profiles.hernquist import Hernquist, HernquistEllipse
from lenstronomy.Util.util import make_grid
import numpy as np
import numpy.testing as npt


class TestHernquist(object):

    def setup_method(self):
        self.hernquist = Hernquist()
        self.hernquist_ellipse = HernquistEllipse()

    def test_total_flux(self):
        delta_pix = 0.2
        x, y = make_grid(numPix=1000, deltapix=delta_pix)

        rs, amp = 1, 1
        total_flux = self.hernquist.total_flux(amp=amp, Rs=rs)
        flux = self.hernquist.function(x, y, amp=amp, Rs=rs)
        total_flux_numerics = np.sum(flux) * delta_pix**2
        npt.assert_almost_equal(total_flux_numerics / total_flux, 1, decimal=1)

        total_flux_ellipse = self.hernquist_ellipse.total_flux(amp=amp, Rs=rs)
        npt.assert_almost_equal(total_flux_ellipse, total_flux)
